## Script (Python) "getUniqueWidgetAttr"
##title=Edit content
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=fields, attr
##

order = []

for f in fields:
    widget = f.widget
    helper = getattr(widget, attr, None)
    # We expect the attribute value to be a iterable.
    if helper:
        # I love list comprehension ;)
        [order.append(item) for item in helper
         if item not in order]

return order



from zope.interface import Interface


class IEditForm(Interface):
    """Archetypes edit form view
    """

    def isTemporaryObject():
        """"""

    def isMultiPageSchema():
        """ Returns true if the schema for the current object should be
        rendered on multiple pages.
        """

    def fieldsets():
        """"""

    def fields(fieldsets):
        """"""

    def getTranslatedSchemaLabel(schema):
        """ Returns the translated title for the given schema. """

    def normalizeString(text):
        """"""



## Script (Python) "session_save_form"
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=
##title=Introspect context schema and saves on session REQUESTs values
##

pass



from zope.interface import Interface, Attribute


class IField(Interface):
    """ Interface for fields """

#     required = Attribute('required', 'Require a value to be present when submitting the field')
#     default = Attribute('default', 'Default value for a field')
#     default_method = Attribute('default_method', 'Name of a method on the field ' +
#                                                  'or an actual function instance to return the default')
#     vocabulary = Attribute('vocabulary', 'List of suggested values for the field')
#     vocabulary_factory = Attribute('vocabulary_factory', 'Name of Zope 3 vocabulary factory')
#     enforceVocabulary = Attribute('enforceVocabulary', \
#                                   'Restrict the allowed values to the ones in the vocabulary')
#     multiValued = Attribute('multiValued', 'Allow the field to have multiple values')
#     searchable = Attribute('searchable', 'Make the field searchable')
#     isMetadata = Attribute('isMetadata', 'Is this field a metadata field?')
#     accessor = Attribute('accessor', 'Use this method as the accessor for the field')
#     mutator = Attribute('mutator', 'Use this method as the mutator for the field')
#     mode = Attribute('mode', 'Mode of access to this field')
#     read_permission = Attribute('read_permission', \
#                                 'Permission to use to protect field reading')
#     write_permission = Attribute('write_permission', \
#                                  'Permission to use to protect writing to the field')

#     storage = Attribute('storage', 'Storage class to use for this field')
#     form_info = Attribute('form_info', 'Form Info (?)')
#     generateMode = Attribute('generateMode', 'Generate Mode (?)')
#     force = Attribute('force', 'Force (?)')
#     type = Attribute('type', 'Type of the field')

    def Vocabulary(content_instance=None):
        """
        returns a DisplayList

        uses self.vocabulary as source

        1) Dynamic vocabulary:

            precondition: a content_instance is given.

            has to return a:
                * DisplayList or
                * list of strings or
                * list of 2-tuples with strings:
                    '[("key1","value 1"),("key 2","value 2"),]'

            the output is postprocessed like a static vocabulary.

            vocabulary is a string:
                if a method with the name of the string exists it will be called

            vocabulary is a class implementing IVocabulary:
                the "getDisplayList" method of the class will be called.


        2) Static vocabulary

            * is already a DisplayList
            * is a list of 2-tuples with strings (see above)
            * is a list of strings (in this case a DisplayList with key=value
              will be created)

        3) Zope 3 vocabulary factory vocabulary

            - precondition: a content_instance is given

            - self.vocabulary_factory is given

            - a named utility providing zope.schema.interfaces.IVocbularyFactory
              exists for the name self.vocabulary_factory.

        """

    # private
    def copy():
        """Return a copy of field instance [PRIVATE]

        Consisting of field name and properties dictionary.
        """

    def validate(value, instance, errors={}, **kwargs):
        """Validate passed-in value using all field validators.

        Return None if all validations pass; otherwise, return failed
        result returned by validator
        """

    # private
    def validate_required(instance, value, errors):
        """Validate the required flag for a field [PRIVATE]

        Overwrite it in your field for special case handling like empty files
        """

    def checkPermission(mode, instance):
        """
        Check whether the security context allows the given permission on
        the given object.

        Arguments:

        permission -- A permission name
        instance -- The object being accessed according to the permission
        """

    def checkExternalEditor(instance):
        """ Checks if the user may edit this field and if
        external editor is enabled on this instance
        """

    def getWidgetName():
        """Return the widget name that is configured for this field as
        a string"""

    def getName():
        """Return the name of this field as a string"""

    def getType():
        """Return the type of this field as a string"""

    def getDefault(instance):
        """Return the default value to be used for initializing this
        field"""

    def getAccessor(instance):
        """Return the accessor method for getting data out of this
        field"""

    def getEditAccessor(instance):
        """Return the accessor method for getting raw data out of this
        field e.g.: for editing
        """

    def getMutator(instance):
        """Return the mutator method used for changing the value
        of this field"""

    # private
    def toString():
        """Utility method for converting a Field to a string [PRIVATE]

        For the purpose of comparing fields.  This comparison is used for
        determining whether a schema has changed in the auto update
        function.  Right now it's pretty crude."""

    def isLanguageIndependent(instance):
        """Get the language independed flag for i18n content (used by LinguaPlon)
        """


class IObjectField(IField):
    """ Interface for fields that support a storage layer """

    # private
    def get(instance, **kwargs):
        """Get the value for this field using the underlying storage [PRIVATE]
        """

    # private
    def getRaw(instance, **kwargs):
        """Get the raw value for this field using the underlying storage [PRIVATE]
        """

    # private
    def set(instance, value, **kwargs):
        """Set the value for this field using the underlying storage [PRIVATE]
        """

    # private
    def unset(instance, **kwargs):
        """Unset the value for this field using the underlying storage [PRIVATE]
        """

    # private
    def getStorage():
        """Return the storage class used in this field [PRIVATE]
        """

    # private
    def setStorage(instance, storage):
        """ Set the storage for this field to the give storage. [PRIVATE]

        Values are migrated by doing a get before changing the storage
        and doing a set after changing the storage.

        The underlying storage must take care of cleaning up of removing
        references to the value stored using the unset method."""

    def getStorageName(instance=None):
        """Return the storage name that is configured for this field
        as a string"""

    def getStorageType(instance=None):
        """Return the type of the storage of this field as a string"""

    def getContentType(instance, fromBaseUnit=True):
        """Return the mime type of object if known or can be guessed;
        otherwise, return None."""

    def get_size(instance):
        """Get size of the stored data used for get_size in BaseObject

        Should be overwritte by special fields like FileField. It's safe for
        fields which are storing strings, ints and BaseUnits but it won't return
        the right results for fields containing OFS.Image.File instances or
        lists/tuples/dicts.
        """


class IFileField(IObjectField):
    """Interface fora fields which (may) contain a file like FileField or
    TextField
    """

    # content_class = Attribute("""Class that is used to wrap the data like
    #                              OFS.Image.File for FileField"""

    # private
    def _process_input(value, default=None, mimetype=None, **kwargs):
        """Processes user input [PRIVATE]

        Must take care of:
            * string
            * open files
            * FileUpload
            * open file like things
            * BaseUnit
            * maybe more ...
        """

    # private
    def getBaseUnit(instance):
        """Return the value of the field wrapped in a base unit object [PRIVATE]
        """

    def getFilename(instance, fromBaseUnit=True):
        """Get file name of underlaying file object
        """

    # private
    def setFilename(instance, filename):
        """Set file name in the base unit [PRIVATE]
        """

    # private
    def download(instance):
        """Kicks download [PRIVATE]

        Writes data including file name and content type to RESPONSE
        """


class IStringField(IField):
    """Marker interface for string fields
    """


class ITextField(IFileField):
    """Marker interface for text fields
    """


class IDateTimeField(IField):
    """Marker interface for date/time field
    """


class ILinesField(IField):
    """Marker interface for lines fields
    """


class IIntegerField(IField):
    """Marker interface for integer fields
    """


class IFloatField(IField):
    """Marker interface for float fields
    """


class IFixedPointField(IField):
    """Marker interface for fixed point fields
    """


class IReferenceField(IField):
    """Marker interface for reference fields
    """


class IComputedField(IField):
    """Marker interface for computed fields
    """


class IBooleanField(IField):
    """Marker interface for boolean fields
    """


class IImageField(IFileField):
    """ Marker interface for detecting an image field """


class IFieldDefaultProvider(Interface):
    """Register a named adapter for your content type providing
    this interface, with a name that is equal to the name of a
    field. If no default or default_method is set on that field
    explicitly, Archetypes will find and call this adapter.
    """

    def __call__():
        """Get the default value.
        """



import sys
from App.class_init import InitializeClass

from Products.Archetypes import PloneMessageFactory as _
from Products.Archetypes.debug import log_exc
from Products.Archetypes.utils import DisplayList
from Products.Archetypes.utils import mapply
from Products.Archetypes.utils import fixSchema
from Products.Archetypes.utils import shasattr
from Products.Archetypes.Field import StringField
from Products.Archetypes.Field import TextField
from Products.Archetypes.Renderer import renderer
from Products.Archetypes.Schema import Schema
from Products.Archetypes.Schema import getSchemata
from Products.Archetypes.Widget import IdWidget
from Products.Archetypes.Widget import StringWidget
from Products.Archetypes.Marshall import RFC822Marshaller
from Products.Archetypes.interfaces import IBaseObject
from Products.Archetypes.interfaces import IReferenceable
from Products.Archetypes.interfaces import IReferenceField
from Products.Archetypes.interfaces import ISchema
from Products.Archetypes.interfaces.field import IFileField
from Products.Archetypes.validator import AttributeValidator
from Products.Archetypes.config import ATTRIBUTE_SECURITY
from Products.Archetypes.config import RENAME_AFTER_CREATION_ATTEMPTS

from Products.Archetypes.event import ObjectInitializedEvent
from Products.Archetypes.event import ObjectEditedEvent

from Products.Archetypes.interfaces import IMultiPageSchema
from Products.Archetypes.interfaces import IObjectPreValidation
from Products.Archetypes.interfaces import IObjectPostValidation

from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from AccessControl.Permissions import copy_or_move as permission_copy_or_move
from Acquisition import aq_base
from Acquisition import aq_inner
from Acquisition import aq_parent
from Acquisition import ImplicitAcquisitionWrapper
from Acquisition import ExplicitAcquisitionWrapper
from Acquisition import Explicit

from ComputedAttribute import ComputedAttribute
from ZODB.POSException import ConflictError
import transaction

from Products.CMFCore import permissions
from Products.CMFCore.utils import getToolByName

from Referenceable import Referenceable

from ZPublisher import xmlrpc
from webdav.NullResource import NullResource

from zope import event
from zope.interface import implements, Interface, providedBy
from zope.component import getSiteManager
from zope.component import subscribers
from zope.component import queryUtility

# Import conditionally, so we don't introduce a hard depdendency
try:
    from plone.i18n.normalizer.interfaces import IUserPreferredURLNormalizer
    from plone.i18n.normalizer.interfaces import IURLNormalizer
    URL_NORMALIZER = True
except ImportError:
    URL_NORMALIZER = False

try:
    from plone.locking.interfaces import ILockable
    HAS_LOCKING = True
except ImportError:
    HAS_LOCKING = False

_marker = []

content_type = Schema((

    StringField(
        name='id',
        required=0,  # Still actually required, but the widget will
                     # supply the missing value on non-submits
        mode='rw',
        permission=permission_copy_or_move,
        accessor='getId',
        mutator='setId',
        default=None,
        widget=IdWidget(
            label=_(u'label_short_name', default=u'Short name'),
            description=_(u'help_shortname',
                          default=u'Should not contain spaces, underscores or mixed case. '
                          'Short Name is part of the item\'s web address.'),
            visible={'view': 'invisible'}
        ),
    ),

    StringField(
        name='title',
        required=1,
        searchable=1,
        default='',
        accessor='Title',
        widget=StringWidget(
            label_msgid='label_title',
            visible={'view': 'invisible'},
            i18n_domain='plone',
        ),
    ),

), marshall=RFC822Marshaller())


class BaseObject(Referenceable):

    security = ClassSecurityInfo()

    # Protect AttributeStorage-based attributes. See the docstring of
    # AttributeValidator for the low-down.
    if ATTRIBUTE_SECURITY:
        attr_security = AttributeValidator()
        security.setDefaultAccess(attr_security)
        # Delete so it cannot be accessed anymore.
        del attr_security

    schema = content_type
    _signature = None

    installMode = ['type', 'actions', 'indexes']

    _at_rename_after_creation = False  # rename object according to title?

    implements(IBaseObject, IReferenceable)

    def __init__(self, oid, **kwargs):
        self.id = oid

    security.declareProtected(permissions.ModifyPortalContent,
                              'initializeArchetype')

    def initializeArchetype(self, **kwargs):
        # Called by the generated add* factory in types tool.
        try:
            self.initializeLayers()
            self.markCreationFlag()
            self.setDefaults()
            if kwargs:
                kwargs['_initializing_'] = True
                self.edit(**kwargs)
            self._signature = self.Schema().signature()
        except (ConflictError, KeyboardInterrupt):
            raise
        except:
            log_exc()

    security.declarePrivate('manage_afterAdd')

    def manage_afterAdd(self, item, container):
        __traceback_info__ = (self, item, container)
        Referenceable.manage_afterAdd(self, item, container)
        self.initializeLayers(item, container)

    security.declarePrivate('manage_afterClone')

    def manage_afterClone(self, item):
        __traceback_info__ = (self, item)
        Referenceable.manage_afterClone(self, item)

    security.declarePrivate('manage_beforeDelete')

    def manage_beforeDelete(self, item, container):
        __traceback_info__ = (self, item, container)
        self.cleanupLayers(item, container)
        Referenceable.manage_beforeDelete(self, item, container)

    security.declarePrivate('initializeLayers')

    def initializeLayers(self, item=None, container=None):
        self.Schema().initializeLayers(self, item, container)

    security.declarePrivate('cleanupLayers')

    def cleanupLayers(self, item=None, container=None):
        self.Schema().cleanupLayers(self, item, container)

    security.declareProtected(permissions.View, 'title_or_id')

    def title_or_id(self):
        """Returns the title if it is not blank and the id otherwise.
        """
        if shasattr(self, 'Title'):
            if callable(self.Title):
                return self.Title() or self.getId()

        return self.getId()

    security.declareProtected(permissions.View, 'getId')

    def getId(self):
        # Gets the object id.
        return self.id

    security.declareProtected(permissions.ModifyPortalContent, 'setId')

    def setId(self, value):
        # Sets the object id.
        # avoid CopyError in OFS.CopySupport.manage_renameObject(),
        # see http://dev.plone.org/ticket/8338
        value = value.strip()
        if value != self.getId():
            parent = aq_parent(aq_inner(self))
            if parent is not None:
                # See Referenceable, keep refs on what is a move/rename
                self._v_cp_refs = 1
                # We can't rename if the object is locked
                if HAS_LOCKING:
                    lockable = ILockable(self)
                    was_locked = False
                    if lockable.locked():
                        was_locked = True
                        lockable.unlock()
                    parent.manage_renameObject(self.id, value)
                    if was_locked:
                        lockable.lock()
                else:
                    parent.manage_renameObject(self.id, value)
            self._setId(value)

    security.declareProtected(permissions.View, 'Type')

    def Type(self):
        # Dublin Core element - Object type.
        #
        # This method is redefined in ExtensibleMetadata but we need this
        # at the object level (i.e. with or without metadata) to interact
        # with the uid catalog.
        if shasattr(self, 'getTypeInfo'):
            ti = self.getTypeInfo()
            if ti is not None:
                return ti.Title()
        return self.meta_type

    security.declareProtected(permissions.View, 'getField')

    def getField(self, key, wrapped=False):
        # Returns a field object.
        return self.Schema().get(key)

    security.declareProtected(permissions.View, 'getWrappedField')

    def getWrappedField(self, key):
        # Gets a field by id which is explicitly wrapped.
        # XXX Maybe we should subclass field from Acquisition.Explicit?
        return ExplicitAcquisitionWrapper(self.getField(key), self)

    security.declareProtected(permissions.View, 'getDefault')

    def getDefault(self, field):
        # Return the default value of a field.
        field = self.getField(field)
        return field.getDefault(self)

    security.declareProtected(permissions.View, 'isBinary')

    def isBinary(self, key):
        # Return wether a field contains binary data.
        field = self.getField(key)
        if IFileField.providedBy(field):
            value = field.getBaseUnit(self)
            return value.isBinary()
        mimetype = self.getContentType(key)
        if mimetype and shasattr(mimetype, 'binary'):
            return mimetype.binary
        elif mimetype and mimetype.find('text') >= 0:
            return 0
        return 1

    security.declareProtected(permissions.View, 'isTransformable')

    def isTransformable(self, name):
        # Returns wether a field is transformable.
        field = self.getField(name)
        return isinstance(field, TextField) or not self.isBinary(name)

    security.declareProtected(permissions.View, 'widget')

    def widget(self, field_name, mode="view", field=None, **kwargs):
        # Returns the rendered widget.
        if field is None:
            field = self.Schema()[field_name]
        widget = field.widget
        return renderer.render(field_name, mode, widget, self, field=field,
                               **kwargs)

    security.declareProtected(permissions.View, 'getFilename')

    def getFilename(self, key=None):
        # Returns the filename from a field.
        value = None

        if key is None:
            field = self.getPrimaryField()
        else:
            field = self.getField(key) or getattr(self, key, None)

        if field and shasattr(field, 'getFilename'):
            return field.getFilename(self)

        return value

    security.declareProtected(permissions.View, 'getContentType')

    def getContentType(self, key=None):
        # Returns the content type from a field.
        value = 'text/plain'

        if key is None:
            field = self.getPrimaryField()
        else:
            field = self.getField(key) or getattr(self, key, None)

        if field and shasattr(field, 'getContentType'):
            return field.getContentType(self)

        return value

    # Backward compatibility
    # Note: ComputedAttribute should never be protected by a security
    # declaration! See http://dev.plone.org/archetypes/ticket/712
    content_type = ComputedAttribute(getContentType, 1)

    # XXX Where's get_content_type comes from??? There's no trace at both
    # Zope and CMF. It should be removed ASAP!
    security.declareProtected(permissions.View, 'get_content_type')
    get_content_type = getContentType

    security.declareProtected(permissions.ModifyPortalContent,
                              'setContentType')

    def setContentType(self, value, key=None):
        # Sets the content type of a field.
        if key is None:
            field = self.getPrimaryField()
        else:
            field = self.getField(key) or getattr(self, key, None)

        if field and IFileField.providedBy(field):
            field.setContentType(self, value)

    security.declareProtected(permissions.ModifyPortalContent, 'setFilename')

    def setFilename(self, value, key=None):
        # Sets the filename of a field.
        if key is None:
            field = self.getPrimaryField()
        else:
            field = self.getField(key) or getattr(self, key, None)

        if field and IFileField.providedBy(field):
            field.setFilename(self, value)

    security.declareProtected(permissions.View, 'getPrimaryField')

    def getPrimaryField(self):
        # The primary field is some object that responds to
        # PUT/manage_FTPget events.
        fields = self.Schema().filterFields(primary=1)
        if fields:
            return fields[0]
        return None

    security.declareProtected(permissions.View, 'get_portal_metadata')

    def get_portal_metadata(self, field):
        # Returns the portal_metadata for a field.
        pmt = getToolByName(self, 'portal_metadata')
        policy = None
        try:
            schema = getattr(pmt, 'DCMI', None)
            spec = schema.getElementSpec(field.accessor)
            policy = spec.getPolicy(self.portal_type)
        except (ConflictError, KeyboardInterrupt):
            raise
        except:
            log_exc()
            return None, False

        if not policy:
            policy = spec.getPolicy(None)

        return DisplayList(map(lambda x: (x, x), policy.allowedVocabulary())), \
            policy.enforceVocabulary()

    security.declareProtected(permissions.View, 'Vocabulary')

    def Vocabulary(self, key):
        # Returns the vocabulary for a specified field.
        vocab, enforce = None, 0
        field = self.getField(key)
        if field:
            if field.isMetadata:
                vocab, enforce = self.get_portal_metadata(field)

            if vocab is None:
                vocab, enforce = field.Vocabulary(self), \
                    field.enforceVocabulary
        if vocab is None:
            vocab = DisplayList()
        return vocab, enforce

    def __getitem__(self, key):
        """Overloads the object's item access.
        """
        # Don't allow key access to hidden attributes
        if key.startswith('_'):
            raise Unauthorized, key

        schema = self.Schema()
        keys = schema.keys()

        if key not in keys and not key.startswith('_'):
            # XXX Fix this in AT 1.4
            value = getattr(aq_inner(self).aq_explicit, key, _marker) or \
                getattr(aq_parent(aq_inner(self)).aq_explicit, key, _marker)
            if value is _marker:
                raise KeyError, key
            else:
                return value

        field = schema[key]
        accessor = field.getEditAccessor(self)
        if not accessor:
            accessor = field.getAccessor(self)

        # This is the access mode used by external editor. We need the
        # handling provided by BaseUnit when its available
        kw = {'raw': 1, 'field': field.__name__}
        value = mapply(accessor, **kw)
        return value

    security.declarePrivate('setDefaults')

    def setDefaults(self):
        # Sets the field values to the default values.
        self.Schema().setDefaults(self)

    security.declareProtected(permissions.ModifyPortalContent, 'update')

    def update(self, **kwargs):
        # Changes the values of the field and reindex the object.
        initializing = kwargs.get('_initializing_', False)
        if initializing:
            del kwargs['_initializing_']
        self.Schema().updateAll(self, **kwargs)
        self._p_changed = 1
        if not initializing:
            # Avoid double indexing during initialization.
            self.reindexObject()

    security.declareProtected(permissions.ModifyPortalContent, 'edit')
    edit = update

    security.declareProtected(permissions.View,
                              'validate_field')

    def validate_field(self, name, value, errors):
        # Field's validate hook.
        #
        # Write a method: validate_foo(new_value) -> "error" or None
        # If there is a validate method defined for a given field invoke
        # it by name
        # name -- the name to register errors under
        # value -- the proposed new value
        # errors -- dict to record errors in
        methodName = "validate_%s" % name
        result = None
        if shasattr(self, methodName):
            method = getattr(self, methodName)
            result = method(value)
            if result is not None:
                errors[name] = result
        return result

    # Pre/post validate hooks that will need to write errors
    # into the errors dict directly using errors[fieldname] = ""
    security.declareProtected(permissions.View, 'pre_validate')

    def pre_validate(self, REQUEST=None, errors=None):
        pass

    security.declareProtected(permissions.View, 'post_validate')

    def post_validate(self, REQUEST=None, errors=None):
        pass

    security.declareProtected(permissions.View, 'validate')

    def validate(self, REQUEST=None, errors=None, data=None, metadata=None):
        # Validates the form data from the request.
        if errors is None:
            errors = {}

        self.pre_validate(REQUEST, errors)

        for pre_validator in subscribers((self,), IObjectPreValidation):
            pre_errors = pre_validator(REQUEST)
            if pre_errors is not None:
                for field_name, error_message in pre_errors.items():
                    if field_name in errors:
                        errors[field_name] += " %s" % error_message
                    else:
                        errors[field_name] = error_message

        if errors:
            return errors
        self.Schema().validate(instance=self, REQUEST=REQUEST,
                               errors=errors, data=data, metadata=metadata)

        self.post_validate(REQUEST, errors)

        for post_validator in subscribers((self,), IObjectPostValidation):
            post_errors = post_validator(REQUEST)
            if post_errors is not None:
                for field_name, error_message in post_errors.items():
                    if field_name in errors:
                        errors[field_name] += " %s" % error_message
                    else:
                        errors[field_name] = error_message

        return errors

    security.declareProtected(permissions.View, 'SearchableText')

    def SearchableText(self):
        # All fields marked as 'searchable' are concatenated together
        # here for indexing purpose.
        data = []
        for field in self.Schema().fields():
            if not field.searchable:
                continue
            if IReferenceField.providedBy(field):
                # waking instances is cheaper than processing a potentially
                # huge vocabulary for getting the title, therefore we handle
                # reference fields seperately
                objs = field.get(self)
                if not isinstance(objs, (list, tuple)):
                    objs = [objs]
                datum = ' '.join([o.Title() for o in objs])
                data.append(datum)

            else:
                method = field.getIndexAccessor(self)
                try:
                    datum = method(mimetype="text/plain")
                except TypeError:
                    # Retry in case typeerror was raised because accessor doesn't
                    # handle the mimetype argument
                    try:
                        datum = method()
                    except (ConflictError, KeyboardInterrupt):
                        raise
                    except:
                        continue
                if datum:
                    vocab = field.Vocabulary(self)
                    if isinstance(datum, (list, tuple)):
                        # Unmangle vocabulary: we index key AND value
                        vocab_values = map(
                            lambda value, vocab=vocab: vocab.getValue(value, ''), datum)
                        vocab_values = [
                            v.encode('utf-8') if isinstance(v, unicode) else v
                            for v in vocab_values
                        ]
                        datum = list(datum)
                        datum.extend(vocab_values)
                        datum = ' '.join(datum)
                    elif isinstance(datum, basestring):
                        if isinstance(datum, unicode):
                            datum = datum.encode('utf-8')
                        value = vocab.getValue(datum, '')
                        if isinstance(value, unicode):
                            value = value.encode('utf-8')
                        datum = "%s %s" % (datum, value, )

                    if isinstance(datum, unicode):
                        datum = datum.encode('utf-8')
                    data.append(str(datum))

        data = ' '.join(data)
        return data

    security.declareProtected(permissions.View, 'getCharset')

    def getCharset(self):
        # Returns the site default charset, or utf-8.
        return 'utf-8'

    security.declareProtected(permissions.View, 'get_size')

    def get_size(self):
        # Used for FTP and apparently the ZMI now too.
        size = 0
        for field in self.Schema().fields():
            size += field.get_size(self)
        return size

    security.declarePrivate('_processForm')

    def _processForm(self, data=1, metadata=None, REQUEST=None, values=None):
        request = REQUEST or self.REQUEST
        if values:
            form = values
        else:
            form = request.form
        fieldset = form.get('fieldset', None)
        schema = self.Schema()
        schemata = self.Schemata()
        fields = []

        if not IMultiPageSchema.providedBy(self):
            fields = schema.fields()
        elif fieldset is not None:
            fields = schemata[fieldset].fields()
        else:
            if data:
                fields += schema.filterFields(isMetadata=0)
            if metadata:
                fields += schema.filterFields(isMetadata=1)

        form_keys = form.keys()

        for field in fields:
            # Delegate to the widget for processing of the form
            # element.  This means that if the widget needs _n_
            # fields under a naming convention it can handle this
            # internally.  The calling API is process_form(instance,
            # field, form) where instance should rarely be needed,
            # field is the field object and form is the dict. of
            # kv_pairs from the REQUEST
            ##
            # The product of the widgets processing should be:
            # (value, **kwargs) which will be passed to the mutator
            # or None which will simply pass

            if not field.writeable(self):
                # If the field has no 'w' in mode, or the user doesn't
                # have the required permission, or the mutator doesn't
                # exist just bail out.
                continue

            try:
                # Pass validating=False to inform the widget that we
                # aren't in the validation phase, IOW, the returned
                # data will be forwarded to the storage
                result = field.widget.process_form(self, field, form,
                                                   empty_marker=_marker,
                                                   validating=False)
            except TypeError:
                # Support for old-style process_form methods
                result = field.widget.process_form(self, field, form,
                                                   empty_marker=_marker)

            if result is _marker or result is None:
                continue

            # Set things by calling the mutator
            mutator = field.getMutator(self)
            __traceback_info__ = (self, field, mutator)
            result[1]['field'] = field.__name__
            mapply(mutator, result[0], **result[1])

        self.reindexObject()

    security.declareProtected(permissions.ModifyPortalContent, 'processForm')

    def processForm(self, data=1, metadata=0, REQUEST=None, values=None):
        # Processes the schema looking for data in the form.
        is_new_object = self.checkCreationFlag()
        self._processForm(data=data, metadata=metadata,
                          REQUEST=REQUEST, values=values)

        if self._at_rename_after_creation and is_new_object:
            self._renameAfterCreation(check_auto_id=True)

        self.unmarkCreationFlag()

        # Post create/edit hooks
        if is_new_object:
            event.notify(ObjectInitializedEvent(self))
            self.at_post_create_script()
        else:
            event.notify(ObjectEditedEvent(self))
            self.at_post_edit_script()

    # This method is only called once after object creation.
    security.declarePrivate('at_post_create_script')

    def at_post_create_script(self):
        pass

    # This method is called after every subsequent edit
    security.declarePrivate('at_post_edit_script')

    def at_post_edit_script(self):
        pass

    security.declareProtected(permissions.ModifyPortalContent,
                              'markCreationFlag')

    def markCreationFlag(self):
        # Sets flag on the instance to indicate that the object hasn't been
        # saved properly (unset in content_edit).
        #
        # This will only be done if a REQUEST is present to ensure that objects
        # created programmatically are considered fully created.
        req = getattr(self, 'REQUEST', None)
        if shasattr(req, 'get'):
            if req.get('SCHEMA_UPDATE', None) is not None:
                return
            meth = req.get('REQUEST_METHOD', None)
            # Ensure that we have an HTTP request, if you're creating an
            # object with something other than a GET or POST, then we assume
            # you are making a complete object.
            if meth in ('GET', 'POST', 'PUT', 'MKCOL'):
                self._at_creation_flag = True

    security.declareProtected(permissions.ModifyPortalContent,
                              'unmarkCreationFlag')

    def unmarkCreationFlag(self):
        # Removes the creation flag.
        if shasattr(aq_inner(self), '_at_creation_flag'):
            self._at_creation_flag = False

    security.declareProtected(permissions.ModifyPortalContent,
                              'checkCreationFlag')

    def checkCreationFlag(self):
        # Returns True if the object has not been fully saved, False otherwise.
        return getattr(aq_base(self), '_at_creation_flag', False)

    def generateNewId(self):
        # Suggest an id for this object.
        # This id is used when automatically renaming an object after creation.
        title = self.Title()
        # Can't work w/o a title
        if not title:
            return None

        # Don't do anything without the plone.i18n package
        if not URL_NORMALIZER:
            return None

        if not isinstance(title, unicode):
            title = unicode(title, 'utf-8')

        request = getattr(self, 'REQUEST', None)
        if request is not None:
            return IUserPreferredURLNormalizer(request).normalize(title)

        return queryUtility(IURLNormalizer).normalize(title)

    security.declarePrivate('_renameAfterCreation')

    def _renameAfterCreation(self, check_auto_id=False):
        # Renames an object like its normalized title.
        old_id = self.getId()
        if check_auto_id and not self._isIDAutoGenerated(old_id):
            # No auto generated id
            return False

        new_id = self.generateNewId()
        if new_id is None:
            return False

        invalid_id = True
        check_id = getattr(self, 'check_id', None)
        if check_id is not None:
            invalid_id = check_id(new_id, required=1)

        # If check_id told us no, or if it was not found, make sure we have an
        # id unique in the parent folder.
        if invalid_id:
            unique_id = self._findUniqueId(new_id)
            if unique_id is not None:
                if check_id is None or check_id(new_id, required=1):
                    new_id = unique_id
                    invalid_id = False

        if not invalid_id:
            # Can't rename without a subtransaction commit when using
            # portal_factory!
            transaction.savepoint(optimistic=True)
            self.setId(new_id)
            return new_id

        return False

    security.declarePrivate('_findUniqueId')

    def _findUniqueId(self, id):
        # Find a unique id in the parent folder, based on the given id, by
        # appending -n, where n is a number between 1 and the constant
        # RENAME_AFTER_CREATION_ATTEMPTS, set in config.py. If no id can be
        # found, return None.
        check_id = getattr(self, 'check_id', None)
        if check_id is None:
            parent = aq_parent(aq_inner(self))
            parent_ids = parent.objectIds()
            check_id = lambda id, required: id in parent_ids

        invalid_id = check_id(id, required=1)
        if not invalid_id:
            return id

        idx = 1
        while idx <= RENAME_AFTER_CREATION_ATTEMPTS:
            new_id = "%s-%d" % (id, idx)
            if not check_id(new_id, required=1):
                return new_id
            idx += 1

        return None

    security.declarePrivate('_isIDAutoGenerated')

    def _isIDAutoGenerated(self, id):
        # Avoid busting setDefaults if we don't have a proper acquisition
        # context.
        plone_tool = getToolByName(self, 'plone_utils', None)
        if plone_tool is not None and \
           shasattr(plone_tool, 'isIDAutoGenerated'):
            return plone_tool.isIDAutoGenerated(id)
        return False

    security.declareProtected(permissions.View, 'Schemata')

    def Schemata(self):
        # Returns the Schemata for the Object.
        return getSchemata(self)

    def Schema(self):
        # Return a (wrapped) schema instance for this object instance.
        return ImplicitAcquisitionWrapper(ISchema(self), self)

    security.declarePrivate('_isSchemaCurrent')

    def _isSchemaCurrent(self):
        # Determines whether the current object's schema is up to date.
        return self._signature == self.Schema().signature()

    security.declarePrivate('_updateSchema')

    def _updateSchema(self, excluded_fields=None, out=None,
                      remove_instance_schemas=False):
        """Updates an object's schema when the class schema changes.

        For each field we use the existing accessor to get its value,
        then we re-initialize the class, then use the new schema
        mutator for each field to set the values again.

        We also copy over any class methods to handle product
        refreshes gracefully (when a product refreshes, you end up
        with both the old version of the class and the new in memory
        at the same time -- you really should restart zope after doing
        a schema update).
        """
        if excluded_fields is None:
            excluded_fields = []
        if out is not None:
            print >> out, 'Updating %s' % (self.getId())

        if remove_instance_schemas and 'schema' in self.__dict__:
            if out is not None:
                print >> out, 'Removing schema from instance dict.'
            del self.schema
        new_schema = self.Schema()

        # Read all the old values into a dict
        values = {}
        mimes = {}
        for f in new_schema.fields():
            name = f.getName()
            if name in excluded_fields:
                continue
            if f.type == "reference":
                continue
            try:
                values[name] = self._migrateGetValue(name, new_schema)
            except ValueError:
                if out is not None:
                    print >> out, ('Unable to get %s.%s'
                                   % (str(self.getId()), name))
            else:
                if shasattr(f, 'getContentType'):
                    mimes[name] = f.getContentType(self)

        obj_class = self.__class__
        current_class = getattr(sys.modules[self.__module__],
                                self.__class__.__name__)
        if obj_class.schema != current_class.schema:
            # XXX This is kind of brutish.  We do this to make sure that old
            # class instances have the proper methods after a refresh.  The
            # best thing to do is to restart Zope after doing an update, and
            # the old versions of the class will disappear.

            for k in current_class.__dict__.keys():
                obj_class.__dict__[k] = current_class.__dict__[k]

        # Set a request variable to avoid resetting the newly created flag
        req = getattr(self, 'REQUEST', None)
        if req is not None:
            req.set('SCHEMA_UPDATE', '1')
        self.initializeArchetype()

        for f in new_schema.fields():
            name = f.getName()
            kw = {}
            if name not in excluded_fields and name in values:
                if name in mimes:
                    kw['mimetype'] = mimes[name]
                try:
                    self._migrateSetValue(name, values[name], **kw)
                except ValueError:
                    if out is not None:
                        print >> out, ('Unable to set %s.%s to '
                                       '%s' % (str(self.getId()),
                                               name, str(values[name])))
        # Make sure the changes are persisted
        self._p_changed = 1
        if out is not None:
            return out

    security.declarePrivate('_migrateGetValue')

    def _migrateGetValue(self, name, new_schema=None):
        # Try to get a value from an object using a variety of methods.
        schema = self.Schema()
        # Migrate pre-AT 1.3 schemas.
        schema = fixSchema(schema)
        # First see if the new field name is managed by the current schema
        field = schema.get(getattr(new_schema.get(
            name, None), 'old_field_name', name), None)
        if field:
            # At very first try to use the BaseUnit itself
            try:
                if IFileField.providedBy(field):
                    return field.getBaseUnit(self)
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

            # First try the edit accessor
            try:
                editAccessor = field.getEditAccessor(self)
                if editAccessor:
                    return editAccessor()
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

            # No luck -- now try the accessor
            try:
                accessor = field.getAccessor(self)
                if accessor:
                    return accessor()
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass
            # No luck use standard method to get the value
            return field.get(self)

            # Still no luck -- try to get the value directly
            # this part should be remove because for some fields this will fail
            # if you get the value directly for example for FixPointField
            # stored value is (0,0) but the input value is a string.
            # at this time FixPointField fails if he got a tuple as input value
            # Because of this line value = value.replace(',','.')
            try:
                return self[field.getName()]
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

        # Nope -- see if the new accessor method is present
        # in the current object.
        if new_schema:
            new_field = new_schema.get(name)
            # Try the new edit accessor
            try:
                editAccessor = new_field.getEditAccessor(self)
                if editAccessor:
                    return editAccessor()
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

            # Nope -- now try the accessor
            try:
                accessor = new_field.getAccessor(self)
                if accessor:
                    return accessor()
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

            # Still no luck -- try to get the value directly using the new name
            try:
                return self[new_field.getName()]
            except (ConflictError, KeyboardInterrupt):
                raise
            except:
                pass

        # Nope -- now see if the current object has an attribute
        # with the same name
        # as the new field
        if shasattr(self, name):
            return getattr(self, name)

        raise ValueError, 'name = %s' % (name)

    security.declarePrivate('_migrateSetValue')

    def _migrateSetValue(self, name, value, old_schema=None, **kw):
        # Try to set an object value using a variety of methods.
        schema = self.Schema()
        # Migrate pre-AT 1.3 schemas.
        schema = fixSchema(schema)
        field = schema.get(name, None)
        # Try using the field's mutator
        if field:
            mutator = field.getMutator(self)
            if mutator is not None:
                try:
                    args = [value]
                    mapply(mutator, *args, **kw)
                    return
                except (ConflictError, KeyboardInterrupt):
                    raise
                except:
                    log_exc()
        else:
            # Try setting an existing attribute
            if shasattr(self, name):
                setattr(self, name, value)
                return
        raise ValueError, 'name = %s, value = %s' % (name, value)

    security.declareProtected(permissions.View, 'isTemporary')

    def isTemporary(self):
        # Checks to see if we are created as temporary object by
        # portal factory.
        parent = aq_parent(aq_inner(self))
        return shasattr(parent, 'meta_type') and \
            parent.meta_type == 'TempFolder'

    security.declareProtected(permissions.View, 'getFolderWhenPortalFactory')

    def getFolderWhenPortalFactory(self):
        # Returns the folder where this object was created temporarily.
        ctx = aq_inner(self)
        if not ctx.isTemporary():
            # Not a temporary object!
            return aq_parent(ctx)
        utool = getToolByName(self, 'portal_url')
        portal_object = utool.getPortalObject()

        while ctx.getId() != 'portal_factory':
            # Find the portal factory object
            if ctx == portal_object:
                # uups, shouldn't happen!
                return ctx
            ctx = aq_parent(ctx)
        # ctx is now the portal_factory in our parent folder
        return aq_parent(ctx)

    #
    # Subobject Access
    #
    # Some temporary objects could be set by fields (for instance
    # additional images that may result from the transformation of
    # a PDF field to html).
    #
    # Those objects are specific to a session.
    #

    security.declareProtected(permissions.ModifyPortalContent,
                              'addSubObjects')

    def addSubObjects(self, objects, REQUEST=None):
        # Adds a dictionary of objects to a volatile attribute.
        if objects:
            storage = getattr(aq_base(self), '_v_at_subobjects', None)
            if storage is None:
                setattr(self, '_v_at_subobjects', {})
                storage = getattr(aq_base(self), '_v_at_subobjects')
            for name, obj in objects.items():
                storage[name] = aq_base(obj)

    security.declareProtected(permissions.View, 'getSubObject')

    def getSubObject(self, name, REQUEST, RESPONSE=None):
        # Gets a dictionary of objects from a volatile attribute.
        storage = getattr(aq_base(self), '_v_at_subobjects', None)
        if storage is None:
            return None

        data = storage.get(name, None)
        if data is None:
            return None

        mtr = self.mimetypes_registry
        mt = mtr.classify(data, filename=name)
        return Wrapper(data, name, str(mt) or 'application/octet-stream').__of__(self)

    def __bobo_traverse__(self, REQUEST, name):
        # Allows transparent access to session subobjects.
        #
        # Sometimes, the request doesn't have a response, e.g. when
        # PageTemplates traverse through the object path, they pass in
        # a phony request (a dict).
        RESPONSE = getattr(REQUEST, 'RESPONSE', None)

        # Is it a registered sub object
        data = self.getSubObject(name, REQUEST, RESPONSE)
        if data is not None:
            return data
        # Or a standard attribute (maybe acquired...)
        target = None
        method = REQUEST.get('REQUEST_METHOD', 'GET').upper()
        # Logic from "ZPublisher.BaseRequest.BaseRequest.traverse"
        # to check whether this is a browser request
        if (len(REQUEST.get('TraversalRequestNameStack', ())) == 0 and
            not (method in ('GET', 'HEAD', 'POST') and not
                 isinstance(RESPONSE, xmlrpc.Response))):
            if shasattr(self, name):
                target = getattr(self, name)
        else:
            if shasattr(self, name):  # attributes of self come first
                target = getattr(self, name)
            else:  # then views
                gsm = getSiteManager()
                factory = gsm.adapters.lookup(
                    (providedBy(self), providedBy(REQUEST)), Interface, name
                )
                if factory is not None:
                    # We don't return the view, we raise an
                    # AttributeError instead (below)
                    target = None
                else:  # then acquired attributes
                    target = getattr(self, name, None)

        if target is not None:
            return target
        elif (method not in ('GET', 'POST') and not
              isinstance(RESPONSE, xmlrpc.Response) and
              REQUEST.maybe_webdav_client):
            return NullResource(self, name, REQUEST).__of__(self)
        else:
            # Raising AttributeError will look up views for us
            raise AttributeError(name)

InitializeClass(BaseObject)


class Wrapper(Explicit):
    """Wrapper object for access to sub objects."""
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, data, filename, mimetype):
        self._data = data
        self._filename = filename
        self._mimetype = mimetype

    def __call__(self, REQUEST=None, RESPONSE=None):
        if RESPONSE is None:
            RESPONSE = REQUEST.RESPONSE
        if RESPONSE is not None:
            mt = self._mimetype
            name = self._filename
            RESPONSE.setHeader('Content-type', str(mt))
            RESPONSE.setHeader('Content-Disposition',
                               'inline;filename=%s' % name)
            RESPONSE.setHeader('Content-Length', len(self._data))
        return self._data

MinimalSchema = BaseObject.schema

__all__ = ('BaseObject', 'MinimalSchema')



import sys
import bbb

from Products.Archetypes.config import PKG_NAME, REGISTER_DEMO_TYPES
from Products.Archetypes.utils import DisplayList

from AccessControl import ModuleSecurityInfo
from AccessControl import allow_class, allow_module
from Products.CMFCore import permissions

###
# security
###
# make log and log_exc public
allow_module('Products.Archetypes.utils')

ModuleSecurityInfo('Products.Archetypes.debug').declarePublic('log')
ModuleSecurityInfo('Products.Archetypes.debug').declarePublic('log_exc')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declarePublic(
    'getAllowableContentTypes')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declarePublic(
    'getAllowedContentTypes')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declarePublic(
    'getForbiddenContentTypes')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declarePublic(
    'getDefaultContentType')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declareProtected(
    permissions.ManagePortal, 'setForbiddenContentTypes')
ModuleSecurityInfo('Products.Archetypes.mimetype_utils').declareProtected(
    permissions.ManagePortal, 'setDefaultContentType')

# Import "PloneMessageFactory as _" to create messages in plone domain
# duplicated here so we don't add a dependency on CMFPlone
from zope.i18nmessageid import MessageFactory
PloneMessageFactory = MessageFactory('plone')
ModuleSecurityInfo('Products.Archetypes').declarePublic('PloneMessageFactory')

# make DisplayList accessible from python scripts and others objects executed
# in a restricted environment
allow_class(DisplayList)

# Allow import of NotFound exception
ModuleSecurityInfo('zExceptions').declarePublic('NotFound')

###
# register tools and content types
###
from Products.Archetypes.ArchetypeTool import process_types, listTypes
from Products.Archetypes.ArchetypeTool import ArchetypeTool
ATToolModule = sys.modules[ArchetypeTool.__module__]  # mpf :|
from Products.Archetypes.ReferenceEngine import ReferenceCatalog as RefTool
from Products.Archetypes.UIDCatalog import UIDCatalog as UIDTool


tools = (ArchetypeTool, RefTool, UIDTool)

types_globals = globals()


def initialize(context):
    from Products.CMFCore import utils

    utils.ToolInit("%s Tool" % PKG_NAME, tools=tools,
                   icon="tool.gif",
                   ).initialize(context)

    if REGISTER_DEMO_TYPES:
        import Products.Archetypes.examples

        content_types, constructors, ftis = process_types(
            listTypes(PKG_NAME), PKG_NAME)

        utils.ContentInit(
            '%s Content' % PKG_NAME,
            content_types=content_types,
            permission=permissions.AddPortalContent,
            extra_constructors=constructors,
            fti=ftis,
        ).initialize(context)
    try:
        from Products.CMFCore.FSFile import FSFile
        from Products.CMFCore.DirectoryView import registerFileExtension
        registerFileExtension('xsl', FSFile)
        registerFileExtension('xul', FSFile)
    except ImportError:
        pass



import sys

from App.class_init import InitializeClass
from AccessControl import ClassSecurityInfo
from Acquisition import ImplicitAcquisitionWrapper

_marker = []


class ArchetypesRenderer:

    security = ClassSecurityInfo()
    # TODO: more security

    def render(self, field_name, mode, widget, instance=None,
               field=None, accessor=None, **kwargs):
        if field is None:
            field = instance.Schema()[field_name]

        if accessor is None:
            accessor = field.getAccessor(instance)

        context = self.setupContext(field_name, mode, widget,
                                    instance, field, accessor, **kwargs)

        result = widget(mode, instance, context)

        del context
        return result

    def setupContext(self, field_name, mode, widget, instance, field,
                     accessor, **kwargs):

        # look for the context in the stack
        frame = sys._getframe()
        context = _marker
        while context is _marker and frame is not None:
            context = frame.f_locals.get('econtext', _marker)
            frame = frame.f_back
        if context is _marker:
            raise RuntimeError, 'Context not found'

        # for editing of multiple AT-based content at once we might want to
        # prefix the field-name.
        if 'fieldprefix' in kwargs:
            field_name = '%s%s' % (kwargs['fieldprefix'], field_name)

        widget = ImplicitAcquisitionWrapper(widget, instance)
        field = ImplicitAcquisitionWrapper(field, instance)
        context.setLocal('here', instance)
        context.setLocal('fieldName', field_name)
        context.setLocal('accessor', accessor)
        context.setLocal('widget', widget)
        context.setLocal('field', field)
        context.setLocal('mode', mode)
        context.setLocal('UID', instance.UID())

        if kwargs:
            for k, v in kwargs.items():
                context.setLocal(k, v)

        del frame
        return context

InitializeClass(ArchetypesRenderer)

renderer = ArchetypesRenderer()



## Script (Python) "at_isEditable"
##title=Check if an at object is editable
##bind container=container
##bind context=context
##bind namespace=
##bind script=script
##bind subpath=traverse_subpath
##parameters=fields

# the script is assuming that an object is editable if the given field list is
# not empty
if len(fields) == 0:
    from AccessControl import Unauthorized
    raise Unauthorized, context



##########################################################################
#
# Copyright (c) 2002-2005, Benjamin Saller <bcsaller@ideasuite.com>, and
#                              the respective authors. All rights reserved.
# For a list of Archetypes contributors see docs/CREDITS.txt.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the author nor the names of its contributors may be used
#   to endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##########################################################################
"""
"""


from Products.Archetypes.tests.attestcase import ATTestCase
from Products.Archetypes.utils import OrderedDict


class OrderedDictTest(ATTestCase):

    def afterSetUp(self):
        self.d = OrderedDict()
        self.d['a'] = '1'
        self.d['b'] = '2'
        self.d['c'] = '3'

    def test_order(self):
        self.assertTrue(self.d.keys() == ['a', 'b', 'c'])
        self.assertTrue(self.d.values() == ['1', '2', '3'])
        self.assertTrue(self.d.items() == [('a', '1'), ('b', '2'), ('c', '3')])
        self.assertTrue(self.d.reverse() == [
                        ('c', '3'), ('b', '2'), ('a', '1')])

    def test_setitem(self):
        self.d['d'] = '4'
        self.assertTrue(self.d.keys() == ['a', 'b', 'c', 'd'])
        self.assertTrue(self.d.values() == ['1', '2', '3', '4'])
        self.assertTrue(self.d.items() == [
                        ('a', '1'), ('b', '2'), ('c', '3'), ('d', '4')])
        self.assertTrue(self.d.reverse() == [
                        ('d', '4'), ('c', '3'), ('b', '2'), ('a', '1')])

        self.d['c'] = 'C'
        self.assertTrue(self.d.keys() == ['a', 'b', 'c', 'd'])
        self.assertTrue(self.d.values() == ['1', '2', 'C', '4'])
        self.assertTrue(self.d.items() == [
                        ('a', '1'), ('b', '2'), ('c', 'C'), ('d', '4')])
        self.assertTrue(self.d.reverse() == [
                        ('d', '4'), ('c', 'C'), ('b', '2'), ('a', '1')])

    def test_del(self):
        del self.d['b']
        self.assertTrue(self.d.keys() == ['a', 'c'])
        self.assertTrue(self.d.values() == ['1', '3'])
        self.assertTrue(self.d.items() == [('a', '1'), ('c', '3')])
        self.assertTrue(self.d.reverse() == [('c', '3'), ('a', '1')])
        self.assertTrue('b' not in self.d)
        self.assertTrue(self.d.get('b', None) == None)

    def test_clear(self):
        self.d.clear()
        self.assertTrue(self.d.keys() == [])
        self.assertTrue(self.d.values() == [])
        self.assertTrue(self.d.items() == [])
        self.assertTrue(self.d.reverse() == [])

    def test_update(self):
        d2 = {'b': 'B', 'd': '4'}
        self.d.update(d2)
        self.assertTrue(self.d.keys() == ['a', 'b', 'c', 'd'])
        self.assertTrue(self.d.values() == ['1', 'B', '3', '4'])
        self.assertTrue(self.d.items() == [
                        ('a', '1'), ('b', 'B'), ('c', '3'), ('d', '4')])
        self.assertTrue(self.d.reverse() == [
                        ('d', '4'), ('c', '3'), ('b', 'B'), ('a', '1')])

    def test_popitem(self):
        (k, v) = self.d.popitem()
        self.assertTrue(k == 'c')
        self.assertTrue(v == '3')


def test_suite():
    from unittest import TestSuite, makeSuite
    suite = TestSuite()
    suite.addTest(makeSuite(OrderedDictTest))
    return suite



class ObjectFieldException(Exception):
    pass


class TextFieldException(ObjectFieldException):
    pass


class FileFieldException(ObjectFieldException):
    pass


class ReferenceException(Exception):
    pass


class SchemaException(Exception):
    pass



from Products.Archetypes.atapi import *
from Products.Archetypes.Marshall import PrimaryFieldMarshaller
from Products.Archetypes.config import PKG_NAME

schema = BaseSchema + Schema((
    FileField('body',
              required=1,
              primary=1,
              widget=FileWidget(),
              ),
),
    marshall=PrimaryFieldMarshaller())


class SimpleFile(BaseContent):
    """An File (test) type"""
    schema = schema
    archetype_name = "Simple File Type"

    def manage_afterPUT(self, data, marshall_data, file, context, mimetype,
                        filename, REQUEST, RESPONSE):
        """For unit tests
        """
        self.called_afterPUT_hook = True

registerType(SimpleFile, PKG_NAME)



from zope.interface import implements

from plone.uuid.interfaces import IUUID

from Products.Archetypes import config
from Products.Archetypes.exceptions import ReferenceException
from Products.Archetypes.interfaces import IReferenceable
from Products.Archetypes.utils import shasattr, isFactoryContained

from Acquisition import aq_base, aq_parent, aq_inner
from OFS.ObjectManager import BeforeDeleteException

from Products.CMFCore.utils import getToolByName
from OFS.CopySupport import CopySource
from OFS.Folder import Folder
from utils import getRelURL

from App.class_init import InitializeClass
from AccessControl import ClassSecurityInfo

####
# In the case of:
# - a copy:
# * we want to lose refs on the new object
# * we want to keep refs on the orig object
# - a cut/paste
# * we want to keep refs
# - a delete:
# * to lose refs
####


class Referenceable(CopySource):
    """ A Mix-in for Referenceable objects """
    isReferenceable = 1
    implements(IReferenceable)

    security = ClassSecurityInfo()
    # Note: methods of this class are made non-publishable by not giving them
    # docstrings.

    def reference_url(self):
        # like absoluteURL, but return a link to the object with this UID"""
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.reference_url(self)

    def hasRelationshipTo(self, target, relationship=None):
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.hasRelationshipTo(self, target, relationship)

    def addReference(self, object, relationship=None, referenceClass=None,
                     updateReferences=True, **kwargs):
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.addReference(self, object, relationship, referenceClass,
                                 updateReferences, **kwargs)

    def deleteReference(self, target, relationship=None):
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.deleteReference(self, target, relationship)

    def deleteReferences(self, relationship=None):
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.deleteReferences(self, relationship)

    def getRelationships(self):
        # What kinds of relationships does this object have
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.getRelationships(self)

    def getBRelationships(self):
        # What kinds of relationships does this object have from others
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        return tool.getBackRelationships(self)

    def getRefs(self, relationship=None, targetObject=None):
        # get all the referenced objects for this object
        tool = getToolByName(self, 'reference_catalog')
        brains = tool.getReferences(self, relationship, targetObject=targetObject,
                                    objects=False)
        if brains:
            return [self._optimizedGetObject(b.targetUID) for b in brains]
        return []

    def _getURL(self):
        # the url used as the relative path based uid in the catalogs
        return getRelURL(self, self.getPhysicalPath())

    def getBRefs(self, relationship=None, targetObject=None):
        # get all the back referenced objects for this object
        tool = getToolByName(self, 'reference_catalog')
        brains = tool.getBackReferences(self, relationship,
                                        targetObject=targetObject, objects=False)
        if brains:
            return [self._optimizedGetObject(b.sourceUID) for b in brains]
        return []

    # aliases
    getReferences = getRefs
    getBackReferences = getBRefs

    def getReferenceImpl(self, relationship=None, targetObject=None):
        # get all the reference objects for this object
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        refs = tool.getReferences(
            self, relationship, targetObject=targetObject)
        if refs:
            return refs
        return []

    def getBackReferenceImpl(self, relationship=None, targetObject=None):
        # get all the back reference objects for this object
        tool = getToolByName(self, config.REFERENCE_CATALOG)
        refs = tool.getBackReferences(
            self, relationship, targetObject=targetObject)
        if refs:
            return refs
        return []

    def _optimizedGetObject(self, uid):
        tool = getToolByName(self, 'uid_catalog', None)
        if tool is None:  # pragma: no cover
            return ''
        tool = aq_inner(tool)
        traverse = aq_parent(tool).unrestrictedTraverse

        _catalog = tool._catalog
        rids = _catalog.indexes['UID']._index.get(uid, ())
        if isinstance(rids, int):
            rids = (rids, )

        for rid in rids:
            path = _catalog.paths[rid]
            obj = traverse(path, default=None)
            if obj is not None:
                return obj

    def _register(self, reference_manager=None):
        # register with the archetype tool for a unique id
        if IUUID(self, None) is not None:
            return

        if reference_manager is None:
            reference_manager = getToolByName(self, config.REFERENCE_CATALOG)
        reference_manager.registerObject(self)

    def _unregister(self):
        # unregister with the archetype tool, remove all references
        reference_manager = getToolByName(self, config.REFERENCE_CATALOG)
        reference_manager.unregisterObject(self)

    def _getReferenceAnnotations(self):
        # given an object, extract the bag of references for which it is the
        # source
        if not getattr(aq_base(self), config.REFERENCE_ANNOTATION, None):
            setattr(self, config.REFERENCE_ANNOTATION,
                    Folder(config.REFERENCE_ANNOTATION))

        return getattr(self, config.REFERENCE_ANNOTATION).__of__(self)

    def _delReferenceAnnotations(self):
        # Removes annotation from self
        if getattr(aq_base(self), config.REFERENCE_ANNOTATION, None):
            delattr(self, config.REFERENCE_ANNOTATION)

    def UID(self):
        return IUUID(self, None)

    def _setUID(self, uid):
        old_uid = IUUID(self, None)
        if old_uid is None:
            # Nothing to be done.
            return
        # Update forward references
        fw_refs = self.getReferenceImpl()
        for ref in fw_refs:
            assert ref.sourceUID == old_uid
            ref.sourceUID = uid
            item = ref
            container = aq_parent(aq_inner(ref))
            # We call manage_afterAdd to inform the
            # reference catalog about changes.
            ref.manage_afterAdd(item, container)
        # Update back references
        back_refs = self.getBackReferenceImpl()
        for ref in back_refs:
            assert ref.targetUID == old_uid
            ref.targetUID = uid
            item = ref
            container = aq_parent(aq_inner(ref))
            # We call manage_afterAdd to inform the
            # reference catalog about changes.
            ref.manage_afterAdd(item, container)
        setattr(self, config.UUID_ATTR, uid)
        item = self
        container = aq_parent(aq_inner(item))
        # We call manage_afterAdd to inform the
        # reference catalog about changes.
        self.manage_afterAdd(item, container)

    def _updateCatalog(self, container):
        # Update catalog after copy, rename ...

        # the UID index needs to be updated for any annotations we
        # carry
        try:
            uc = getToolByName(container, config.UID_CATALOG)
        except AttributeError:
            # TODO when trying to rename or copy a whole site than
            # container is the object "under" the portal so we can
            # NEVER ever find the catalog which is bad ...
            container = aq_parent(self)
            uc = getToolByName(container, config.UID_CATALOG)

        rc = getToolByName(uc, config.REFERENCE_CATALOG)

        self._catalogUID(container, uc=uc)
        self._catalogRefs(container, uc=uc, rc=rc)

    # OFS Hooks
    def manage_afterAdd(self, item, container):
        # Get a UID
        # (Called when the object is created or moved.)

        if isFactoryContained(self):
            return
        isCopy = getattr(item, '_v_is_cp', None)
        # Before copying we take a copy of the references that are to be copied
        # on the new copy
        rfields = self.Schema().filterFields(type="reference", keepReferencesOnCopy=1)
        rrefs = {}
        if isCopy:
            # If the object is a copy of a existing object we
            # want to renew the UID, and drop all existing references
            # on the newly-created copy.
            for r in rfields:
                rrefs[r.getName()] = r.get(self)
            setattr(self, config.UUID_ATTR, None)
            self._delReferenceAnnotations()

        ct = getToolByName(container, config.REFERENCE_CATALOG, None)
        self._register(reference_manager=ct)
        self._updateCatalog(container)
        self._referenceApply('manage_afterAdd', item, container)
        # copy the references
        if isCopy:
            for r in rfields:
                r.set(self, rrefs[r.getName()])

    def manage_afterClone(self, item):
        # Get a new UID (effectivly dropping reference)
        # (Called when the object is cloned.)

        uc = getToolByName(self, config.UID_CATALOG)

        isCopy = getattr(item, '_v_is_cp', None)
        if isCopy:
            # if isCopy is True, manage_afterAdd should have assigned a
            # UID already.  Don't mess with UID anymore.
            return

        # TODO Should we ever get here after the isCopy flag addition??
        # If the object has no UID or the UID already exists, then
        # we should get a new one

        uuid = IUUID(self, None)

        if (uuid is None or
                len(uc(UID=uuid))):
            setattr(self, config.UUID_ATTR, None)

        self._register()
        self._updateCatalog(self)

    def manage_beforeDelete(self, item, container):
        # Remove self from the catalog.
        # (Called when the object is deleted or moved.)

        # Change this to be "item", this is the root of this recursive
        # chain and it will be flagged in the correct mode
        storeRefs = getattr(item, '_v_cp_refs', None)
        if storeRefs is None:
            # The object is really going away, we want to remove
            # its references
            rc = getToolByName(self, config.REFERENCE_CATALOG)
            references = rc.getReferences(self)
            back_references = rc.getBackReferences(self)
            try:
                # First check the 'delete cascade' case
                if references:
                    for ref in references:
                        ref.beforeSourceDeleteInformTarget()
                # Then check the 'holding/ref count' case
                if back_references:
                    for ref in back_references:
                        ref.beforeTargetDeleteInformSource()
                # If nothing prevented it, remove all the refs
                self.deleteReferences()
            except ReferenceException, E:
                raise BeforeDeleteException(E)

        self._referenceApply('manage_beforeDelete', item, container)

        # Track the UUID
        # The object has either gone away, moved or is being
        # renamed, we still need to remove all UID/child refs
        self._uncatalogUID(container)
        self._uncatalogRefs(container)

    # Catalog Helper methods
    def _catalogUID(self, aq, uc=None):
        if not uc:
            uc = getToolByName(aq, config.UID_CATALOG)
        url = self._getURL()
        uc.catalog_object(self, url)

    def _uncatalogUID(self, aq, uc=None):
        if isFactoryContained(self):
            return
        if not uc:
            uc = getToolByName(self, config.UID_CATALOG)
        url = self._getURL()
        # XXX This is an ugly workaround. This method shouldn't be called
        # twice for an object in the first place, so we don't have to check
        # if it is still cataloged.
        rid = uc.getrid(url)
        if rid is not None:
            uc.uncatalog_object(url)

    def _catalogRefs(self, aq, uc=None, rc=None):
        annotations = self._getReferenceAnnotations()
        if annotations:
            if not uc:
                uc = getToolByName(aq, config.UID_CATALOG)
            if not rc:
                rc = getToolByName(aq, config.REFERENCE_CATALOG)
            for ref in annotations.objectValues():
                url = getRelURL(uc, ref.getPhysicalPath())
                uc.catalog_object(ref, url)
                rc.catalog_object(ref, url)
                ref._catalogRefs(uc, uc, rc)

    def _uncatalogRefs(self, aq, uc=None, rc=None):
        if isFactoryContained(self):
            return
        annotations = self._getReferenceAnnotations()
        if annotations:
            if not uc:
                uc = getToolByName(self, config.UID_CATALOG)
            if not rc:
                rc = getToolByName(self, config.REFERENCE_CATALOG)
            for ref in annotations.objectValues():
                url = getRelURL(uc, ref.getPhysicalPath())
                # XXX This is an ugly workaround. This method shouldn't be
                # called twice for an object in the first place, so we don't
                # have to check if it is still cataloged.
                uc_rid = uc.getrid(url)
                if uc_rid is not None:
                    uc.uncatalog_object(url)
                rc_rid = rc.getrid(url)
                if rc_rid is not None:
                    rc.uncatalog_object(url)

    def _getCopy(self, container):
        # We only set the '_v_is_cp' flag here if it was already set.
        #
        # _getCopy gets called after _notifyOfCopyTo, which should set
        # _v_cp_refs appropriatedly.
        #
        # _getCopy is also called from WebDAV MOVE (though not from
        # 'manage_pasteObjects')
        is_cp_flag = getattr(self, '_v_is_cp', None)
        cp_refs_flag = getattr(self, '_v_cp_refs', None)
        ob = CopySource._getCopy(self, container)
        if is_cp_flag:
            setattr(ob, '_v_is_cp', is_cp_flag)
        if cp_refs_flag:
            setattr(ob, '_v_cp_refs', cp_refs_flag)
        return ob

    def _notifyOfCopyTo(self, container, op=0):
        """keep reference info internally when op == 1 (move)
        because in those cases we need to keep refs"""
        # This isn't really safe for concurrent usage, but the
        # worse case is not that bad and could be fixed with a reindex
        # on the archetype tool
        if op == 1:
            self._v_cp_refs = 1
            self._v_is_cp = 0
        if op == 0:
            self._v_cp_refs = 0
            self._v_is_cp = 1

    # Recursion Mgmt
    def _referenceApply(self, methodName, *args, **kwargs):
        # We always apply commands to our reference children
        # and if we are folderish we need to get those too
        # where as references are concerned
        children = []
        if shasattr(self, 'objectValues'):
            # Only apply to objects that subclass
            # from Referenceable, and only apply the
            # method from Referenceable. Otherwise manage_* will get
            # called multiple times.
            nc = lambda obj: isinstance(obj, Referenceable)
            children.extend(filter(nc, self.objectValues()))
        children.extend(self._getReferenceAnnotations().objectValues())
        if children:
            for child in children:
                if shasattr(Referenceable, methodName):
                    method = getattr(Referenceable, methodName)
                    method(*((child,) + args), **kwargs)

InitializeClass(Referenceable)



"""
This file is here to make the module importable.
"""
from annotations import IATAnnotatable, IATAnnotations
from base import IBaseObject, IBaseContent, IBaseFolder, IBaseUnit
from event import IObjectInitializedEvent, IObjectEditedEvent
from event import IEditBegunEvent, IEditCancelledEvent
from event import IWebDAVObjectInitializedEvent, IWebDAVObjectEditedEvent
from Products.Archetypes.interfaces.field import IField
from Products.Archetypes.interfaces.field import IObjectField
from Products.Archetypes.interfaces.field import IStringField
from Products.Archetypes.interfaces.field import ITextField
from Products.Archetypes.interfaces.field import IDateTimeField
from Products.Archetypes.interfaces.field import ILinesField
from Products.Archetypes.interfaces.field import IIntegerField
from Products.Archetypes.interfaces.field import IFloatField
from Products.Archetypes.interfaces.field import IFileField
from Products.Archetypes.interfaces.field import IImageField
from Products.Archetypes.interfaces.field import IFixedPointField
from Products.Archetypes.interfaces.field import IReferenceField
from Products.Archetypes.interfaces.field import IComputedField
from Products.Archetypes.interfaces.field import IBooleanField
from field import IFileField, IFieldDefaultProvider
from layer import ILayer, ILayerContainer, ILayerRuntime
from marshall import IMarshall
from metadata import IExtensibleMetadata
from orderedfolder import IOrderedFolder
from referenceable import IReferenceable
from referenceengine import IReference, IContentReference
from referenceengine import IReferenceCatalog, IUIDCatalog
from schema import ISchema, ISchemata, ICompositeSchema
from schema import IBindableSchema, IManagedSchema, IMultiPageSchema
from storage import IStorage, ISQLStorage
from templatemixin import ITemplateMixin
from vocabulary import IVocabulary
from athistoryaware import IATHistoryAware
from archetypetool import IArchetypeTool
from edit import IEditForm
from validator import IObjectPreValidation, IObjectPostValidation
from viewlet import IEditBeforeFieldsets, IEditAfterFieldsets

import zope.deferredimport
zope.deferredimport.deprecated(
    "Please use the canonical interface from OFS. "
    "This alias will be removed in the next major version.",
    IOrderedContainer='OFS.interfaces:IOrderedContainer',
)



from Products.Archetypes import WebDAVSupport
from Products.Archetypes.CatalogMultiplex import CatalogMultiplex
from Products.Archetypes.ExtensibleMetadata import ExtensibleMetadata
from Products.Archetypes.BaseObject import BaseObject
from Products.Archetypes.event import WebDAVObjectInitializedEvent
from Products.Archetypes.interfaces import IBaseFolder
from Products.Archetypes.interfaces import IBaseObject
from Products.Archetypes.interfaces import IReferenceable
from Products.Archetypes.interfaces import IExtensibleMetadata

from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from App.class_init import InitializeClass
from Products.CMFCore import permissions
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.PortalFolder import PortalFolderBase as PortalFolder
from Products.CMFCore.PortalContent import PortalContent
from Products.CMFCore.utils import _checkPermission

from zope import event
from zope.interface import implements

FOLDER_MANAGE_OPTIONS = (
    {'action': 'manage_main', 'label': 'Contents'},
    {'action': 'manage_access', 'label': 'Security'},
    {'action': 'manage_interfaces', 'label': 'Interfaces'},
)


class BaseFolderMixin(CatalogMultiplex,
                      BaseObject,
                      PortalFolder,
                      ):
    """A not-so-basic Folder implementation, with no Dublin Core
    Metadata"""

    implements(IBaseFolder, IBaseObject, IReferenceable, IContentish)

    security = ClassSecurityInfo()

    # Fix permissions set by CopySupport.py
    __ac_permissions__ = (
        (permissions.ModifyPortalContent,
         ('manage_cutObjects', 'manage_pasteObjects',
          'manage_renameObject', 'manage_renameObjects',)),
    )
    security.declareProtected('Copy or Move', 'manage_copyObjects')

    manage_options = FOLDER_MANAGE_OPTIONS
    content_icon = "folder_icon.gif"
    isPrincipiaFolderish = 1
    isAnObjectManager = 1
    __dav_marshall__ = False

    __call__ = PortalContent.__call__.im_func

    # This special value informs ZPublisher to use __call__
    index_html = None

    def __init__(self, oid, **kwargs):
        # Call skinned first cause baseobject will set new defaults on
        # those attributes anyway
        PortalFolder.__init__(self, oid, self.Title())
        BaseObject.__init__(self, oid, **kwargs)

    def _notifyOfCopyTo(self, container, op=0):
        """In the case of a move (op=1) we need to make sure
        references are mainained for all referencable objects within
        the one being moved.

        manage_renameObject calls _notifyOfCopyTo so that the
        object being renamed doesn't lose its references. But
        manage_renameObject calls _delObject which calls
        manage_beforeDelete on all the children of the object
        being renamed which deletes all references for children
        of the object being renamed. Here is a patch that does
        recursive calls for _notifyOfCopyTo to address that
        problem.
        """
        # XXX this doesn't appear to be necessary anymore, if it is
        # it needs to be used in BaseBTreeFolder as well, it currently
        # is not.
        BaseObject._notifyOfCopyTo(self, container, op=op)
        # keep reference info internally when op == 1 (move)
        # because in those cases we need to keep refs
        if op == 1:
            self._v_cp_refs = 1
        for child in self.contentValues():
            if IReferenceable.providedBy(child):
                child._notifyOfCopyTo(self, op)

    security.declarePrivate('manage_afterAdd')

    def manage_afterAdd(self, item, container):
        BaseObject.manage_afterAdd(self, item, container)

    security.declarePrivate('manage_afterClone')

    def manage_afterClone(self, item):
        BaseObject.manage_afterClone(self, item)

    security.declarePrivate('manage_beforeDelete')

    def manage_beforeDelete(self, item, container):
        BaseObject.manage_beforeDelete(self, item, container)
        # and reset the rename flag (set in Referenceable._notifyCopyOfCopyTo)
        self._v_cp_refs = None

    security.declareProtected(permissions.DeleteObjects,
                              'manage_delObjects')

    def manage_delObjects(self, ids=None, REQUEST=None):
        """We need to enforce security."""
        if ids is None:
            ids = []
        if isinstance(ids, basestring):
            ids = [ids]
        for id in ids:
            item = self._getOb(id)
            if not _checkPermission(permissions.DeleteObjects, item):
                raise Unauthorized, (
                    "Do not have permissions to remove this object")
        return PortalFolder.manage_delObjects(self, ids, REQUEST=REQUEST)

    security.declareProtected(permissions.ListFolderContents,
                              'listFolderContents')

    def listFolderContents(self, contentFilter=None, suppressHiddenFiles=0):
        # Optionally you can suppress "hidden" files, or files that begin
        # with a dot.
        contents = PortalFolder.listFolderContents(
            self, contentFilter=contentFilter)
        if suppressHiddenFiles:
            contents = [obj for obj in contents if obj.getId()[:1] != '.']

        return contents

    security.declareProtected(permissions.AccessContentsInformation,
                              'folderlistingFolderContents')

    def folderlistingFolderContents(self, contentFilter=None,
                                    suppressHiddenFiles=0):
        # Calls listFolderContents in protected only by ACI so that
        # folder_listing can work without the List folder contents permission.
        return self.listFolderContents(contentFilter=contentFilter,
                                       suppressHiddenFiles=suppressHiddenFiles)

    security.declareProtected(permissions.View, 'Title')

    def Title(self, **kwargs):
        # We have to override Title here to handle arbitrary arguments since
        # PortalFolder defines it.
        return self.getField('title').get(self, **kwargs)

    security.declareProtected(permissions.ModifyPortalContent,
                              'setTitle')

    def setTitle(self, value, **kwargs):
        # We have to override setTitle here to handle arbitrary
        # arguments since PortalFolder defines it.
        self.getField('title').set(self, value, **kwargs)

    def __getitem__(self, key):
        """Overwrite __getitem__.

        At first it's using the BaseObject version. If the element can't be
        retrieved from the schema it's using PortalFolder as fallback which
        should be the ObjectManager's version.
        """
        try:
            return BaseObject.__getitem__(self, key)
        except KeyError:
            return PortalFolder.__getitem__(self, key)

    # override "CMFCore.PortalFolder.PortalFolder.manage_addFolder"
    # as it insists on creating folders of type "Folder".
    # use instead "_at_type_subfolder" or our own type.
    security.declareProtected(permissions.AddPortalFolders,
                              'manage_addFolder')

    def manage_addFolder(self,
                         id,
                         title='',
                         REQUEST=None,
                         type_name=None):
        """Add a new folder-like object with id *id*.

        IF present, use the parent object's 'mkdir' alias; otherwise, just add
        a PortalFolder.
        """
        ti = self.getTypeInfo()
        # BBB getMethodURL is part of CMF 1.5 but AT 1.3 should be compatible
        # with CMF 1.4
        try:
            method = ti and ti.getMethodURL('mkdir') or None
        except AttributeError:
            method = None
        if method is not None:
            # call it
            getattr(self, method)(id=id)
        else:
            if type_name is None:
                type_name = getattr(self, '_at_type_subfolder', None)
            if type_name is None:
                type_name = ti and ti.getId() or 'Folder'
            self.invokeFactory(type_name, id=id)

        ob = self._getOb(id)
        try:
            ob.setTitle(title)
        except AttributeError:
            pass
        try:
            ob.reindexObject()
        except AttributeError:
            pass

    def MKCOL_handler(self, id, REQUEST=None, RESPONSE=None):
        """Hook into the MKCOL (make collection) process to call
        manage_afterMKCOL.
        """
        result = PortalFolder.MKCOL_handler(self, id, REQUEST, RESPONSE)
        new_folder = self._getOb(id)
        if new_folder is not None:  # Could it have been renamed?
            new_folder.unmarkCreationFlag()
            event.notify(WebDAVObjectInitializedEvent(new_folder))
        self.manage_afterMKCOL(id, result, REQUEST, RESPONSE)
        return result

    security.declarePrivate('manage_afterMKCOL')

    def manage_afterMKCOL(self, id, result, REQUEST=None, RESPONSE=None):
        # After MKCOL handler.
        pass

    security.declareProtected(permissions.ModifyPortalContent, 'PUT')
    PUT = WebDAVSupport.PUT

    security.declareProtected(permissions.View, 'manage_FTPget')
    manage_FTPget = WebDAVSupport.manage_FTPget

    security.declarePrivate('manage_afterPUT')
    manage_afterPUT = WebDAVSupport.manage_afterPUT

InitializeClass(BaseFolderMixin)


class BaseFolder(BaseFolderMixin, ExtensibleMetadata):
    """A not-so-basic Folder implementation, with Dublin Core
    Metadata included"""

    implements(IBaseFolder, IBaseObject, IReferenceable,
               IContentish, IExtensibleMetadata)

    schema = BaseFolderMixin.schema + ExtensibleMetadata.schema

    security = ClassSecurityInfo()

    def __init__(self, oid, **kwargs):
        # Call skinned first cause baseobject will set new defaults on
        # those attributes anyway
        BaseFolderMixin.__init__(self, oid, **kwargs)
        ExtensibleMetadata.__init__(self)

    security.declareProtected(permissions.View,
                              'Description')

    def Description(self, **kwargs):
        # We have to override Description here to handle arbitrary
        # arguments since PortalFolder defines it.
        return self.getField('description').get(self, **kwargs)

    security.declareProtected(permissions.ModifyPortalContent,
                              'setDescription')

    def setDescription(self, value, **kwargs):
        # We have to override setDescription here to handle arbitrary
        # arguments since PortalFolder defines it.
        self.getField('description').set(self, value, **kwargs)

InitializeClass(BaseFolder)

BaseFolderSchema = BaseFolder.schema

__all__ = ('BaseFolder', 'BaseFolderMixin', 'BaseFolderSchema')



import string
from logging import DEBUG
from zope.component import queryUtility
from zope.interface import implements

from Products.Archetypes import PloneMessageFactory as _
from Products.Archetypes.Field import BooleanField, LinesField, TextField, \
    StringField, DateTimeField
from Products.Archetypes.Widget import (
    BooleanWidget, TagsWidget, TextAreaWidget, StringWidget,
    DatetimeWidget, SelectWidget, AjaxSelectWidget)
from Products.Archetypes.Schema import Schema
from Products.Archetypes.Schema import MetadataSchema
from Products.Archetypes.interfaces import IExtensibleMetadata
from Products.Archetypes.log import log
from Products.Archetypes.utils import DisplayList, shasattr
from Products.Archetypes import config
import Persistence
from Acquisition import aq_base
from AccessControl import ClassSecurityInfo
from AccessControl import Unauthorized
from DateTime.DateTime import DateTime
from App.class_init import InitializeClass
from App.special_dtml import DTMLFile
from Products.CMFCore import permissions
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.log import log_deprecated
from ComputedAttribute import ComputedAttribute

_marker = []

# http://www.zope.org/Collectors/CMF/325
# http://www.zope.org/Collectors/CMF/476
_zone = DateTime().timezone()

FLOOR_DATE = DateTime(1000, 1)  # always effective
CEILING_DATE = DateTime(2500, 0)  # never expires

# We import this conditionally, in order not to introduce a hard dependency
try:
    from plone.i18n.locales.interfaces import IMetadataLanguageAvailability
    HAS_PLONE_I18N = True
except ImportError:
    HAS_PLONE_I18N = False


# MIXIN
class ExtensibleMetadata(Persistence.Persistent):
    """ A DC metadata implementation for Plone Archetypes
    """

    implements(IExtensibleMetadata)

    security = ClassSecurityInfo()

    schema = type = MetadataSchema((
        BooleanField(
            'allowDiscussion',
            accessor="isDiscussable",
            mutator="allowDiscussion",
            edit_accessor="editIsDiscussable",
            default=None,
            enforceVocabulary=1,
            widget=BooleanWidget(
                label=_(u'label_allow_comments',
                        default=u'Allow comments'),
                description=_(u'help_allow_comments',
                              default=u'If selected, users can add comments '
                              'to this item.')
            ),
        ),
        LinesField(
            'subject',
            multiValued=1,
            accessor="Subject",
            searchable=True,
            widget=TagsWidget(
                label=_(u'label_tags', default=u'Tags'),
                description=_(u'help_tags',
                              default=u'Tags are commonly used for ad-hoc '
                              'organization of content.'),
            ),
        ),
        TextField(
            'description',
            default='',
            searchable=1,
            accessor="Description",
            default_content_type='text/plain',
            allowable_content_types=('text/plain',),
            widget=TextAreaWidget(
                label=_(u'label_description', default=u'Description'),
                description=_(u'help_description',
                              default=u'Used in item listings and search results.'),
            ),
        ),
        # Location, also known as Coverage in the DC metadata standard, but we
        # keep the term Location here for historical reasons.
        StringField(
            'location',
            # why no accessor? http://dev.plone.org/plone/ticket/6424
            searchable=True,
            widget=StringWidget(
                label=_(u'label_location', default=u'Location'),
                description=_(u'help_location_dc',
                              default=u'The geographical location associated with the item, if applicable.'),
            ),
        ),
        LinesField(
            'contributors',
            accessor="Contributors",
            widget=AjaxSelectWidget(
                label=_(u'label_contributors', u'Contributors'),
                description=_(u'help_contributors',
                              default=u"The names of people that have contributed "
                              "to this item. Each contributor should "
                              "be on a separate line."),
                vocabulary="plone.app.vocabularies.Users"
            ),
        ),
        LinesField(
            'creators',
            accessor="Creators",
            widget=AjaxSelectWidget(
                label=_(u'label_creators', u'Creators'),
                description=_(u'help_creators',
                              default=u"Persons responsible for creating the content of "
                              "this item. Please enter a list of user names, one "
                              "per line. The principal creator should come first."),
                vocabulary="plone.app.vocabularies.Users"
            ),
        ),
        DateTimeField(
            'effectiveDate',
            mutator='setEffectiveDate',
            languageIndependent=True,
            widget=DatetimeWidget(
                label=_(u'label_effective_date', u'Publishing Date'),
                description=_(u'help_effective_date',
                              default=u"The date when the item will be published. If no "
                              "date is selected the item will be published immediately."),
            ),
        ),
        DateTimeField(
            'expirationDate',
            mutator='setExpirationDate',
            languageIndependent=True,
            widget=DatetimeWidget(
                label=_(u'label_expiration_date', u'Expiration Date'),
                description=_(u'help_expiration_date',
                              default=u"The date when the item expires. This will automatically "
                              "make the item invisible for others at the given date. "
                              "If no date is chosen, it will never expire."),
            ),
        ),
        StringField(
            'language',
            accessor="Language",
            default=config.LANGUAGE_DEFAULT,
            default_method='defaultLanguage',
            vocabulary='languages',
            widget=SelectWidget(
                label=_(u'label_language', default=u'Language'),
            ),
        ),
        TextField(
            'rights',
            accessor="Rights",
            default_method='defaultRights',
            allowable_content_types=('text/plain',),
            widget=TextAreaWidget(
                label=_(u'label_copyrights', default=u'Rights'),
                description=_(u'help_copyrights',
                              default=u'Copyright statement or other rights information on this item.'),
            )),
    )) + Schema((
        # XXX change this to MetadataSchema in AT 1.4
        # Currently we want to stay backward compatible without migration
        # between beta versions so creation and modification date are using the
        # standard schema which leads to AttributeStorage
        DateTimeField(
                'creation_date',
                accessor='created',
                mutator='setCreationDate',
                default_method=DateTime,
                languageIndependent=True,
                isMetadata=True,
                schemata='metadata',
                generateMode='mVc',
                widget=DatetimeWidget(
                    label=_(u'label_creation_date', default=u'Creation Date'),
                    description=_(u'help_creation_date',
                                  default=u'Date this object was created'),
                    visible={'edit': 'invisible', 'view': 'invisible'}),
                ),
        DateTimeField(
            'modification_date',
            accessor='modified',
            mutator='setModificationDate',
            default_method=DateTime,
            languageIndependent=True,
            isMetadata=True,
            schemata='metadata',
            generateMode='mVc',
            widget=DatetimeWidget(
                    label=_(u'label_modification_date',
                            default=u'Modification Date'),
                    description=_(u'help_modification_date',
                                  default=u'Date this content was modified last'),
                    visible={'edit': 'invisible', 'view': 'invisible'}),
        ),
    ))

    def __init__(self):
        pass

    security.declarePrivate('defaultLanguage')

    def defaultLanguage(self):
        # Retrieve the default language
        tool = getToolByName(self, 'portal_languages', None)
        if tool is not None:
            return tool.getDefaultLanguage()
        return config.LANGUAGE_DEFAULT

    security.declarePrivate('defaultRights')

    def defaultRights(self):
        # Retrieve the default rights.
        mdtool = getToolByName(self, 'portal_metadata', None)
        if mdtool is None:
            return ''
        for sid, schema in mdtool.listSchemas():
            if not hasattr(schema, 'listPolicies'):
                # Broken class from CMFDefault.
                continue
            for pid, policy in schema.listPolicies(typ=self.Type()):
                if pid != 'Rights' and not policy.supply_default:
                    continue
                return policy.default_value
        return ''

    security.declareProtected(permissions.View, 'isDiscussable')

    def isDiscussable(self, encoding=None):
        log_deprecated(
            "The isDiscussable method from the ExtensibleMetadata in "
            "Products.ATContentTypes has been deprecated and will be removed "
            "in Plone 5. This method belongs to the old discussion "
            "infrastructure that already has been replaced by "
            "plone.app.discussion in Plone 4.1."
        )
        if not 'portal_discussion' in self.objectIds():
            return
        # Returns either True or False
        dtool = getToolByName(self, 'portal_discussion')
        return dtool.isDiscussionAllowedFor(self)

    security.declareProtected(permissions.View, 'editIsDiscussable')

    def editIsDiscussable(self, encoding=None):
        log_deprecated(
            "The editIsDiscussable method from the ExtensibleMetadata in "
            "Products.ATContentTypes has been deprecated and will be removed "
            "in Plone 5. This method belongs to the old discussion "
            "infrastructure that already has been replaced by "
            "plone.app.discussion in Plone 4.1."
        )
        if not 'portal_discussion' in self.objectIds():
            return
        # Returns True, False or if None the default value
        result = self.rawIsDiscussable()
        if result is not None:
            return result
        default = self.defaultIsDiscussable()
        return default

    security.declareProtected(permissions.View, 'rawIsDiscussable')

    def rawIsDiscussable(self):
        log_deprecated(
            "The rawIsDiscussable method from the ExtensibleMetadata in "
            "Products.ATContentTypes has been deprecated and will be removed "
            "in Plone 5. This method belongs to the old discussion "
            "infrastructure that already has been replaced by "
            "plone.app.discussion in Plone 4.1."
        )
        if not 'portal_discussion' in self.objectIds():
            return
        # Returns True, False or None where None means use the default
        result = getattr(aq_base(self), 'allow_discussion', None)
        if result is not None:
            result = bool(result)
        return result

    security.declareProtected(permissions.View, 'defaultIsDiscussable')

    def defaultIsDiscussable(self):
        log_deprecated(
            "The defaultIsDiscussable method from the ExtensibleMetadata in "
            "Products.ATContentTypes has been deprecated and will be removed "
            "in Plone 5. This method belongs to the old discussion "
            "infrastructure that already has been replaced by "
            "plone.app.discussion in Plone 4.1."
        )
        if not 'portal_discussion' in self.objectIds():
            return
        # Returns the default value, either True or False
        default = None
        typeInfo = self.getTypeInfo()
        if typeInfo:
            default = typeInfo.allowDiscussion()
        return default

    security.declareProtected(permissions.ModifyPortalContent,
                              'allowDiscussion')

    def allowDiscussion(self, allowDiscussion=None, **kw):
        pass

    # Vocabulary methods ######################################################

    security.declareProtected(permissions.View, 'languages')

    def languages(self):
        # Vocabulary method for the language field.
        util = None

        use_combined = False
        # Respect the combined language code setting from PloneLanguageTool
        lt = getToolByName(self, 'portal_languages', None)
        if lt is not None:
            use_combined = lt.use_combined_language_codes

        # Try the utility first
        if HAS_PLONE_I18N:
            util = queryUtility(IMetadataLanguageAvailability)
        # Fall back to acquiring availableLanguages
        if util is None:
            languages = getattr(self, 'availableLanguages', None)
            if callable(languages):
                languages = languages()
            # Fall back to static definition
            if languages is None:
                return DisplayList(
                    (('en', 'English'), ('fr', 'French'), ('es', 'Spanish'),
                     ('pt', 'Portuguese'), ('ru', 'Russian')))
        else:
            languages = util.getLanguageListing(combined=use_combined)
            languages.sort(key=lambda x: x[1])
            # Put language neutral at the top.
            languages.insert(0, (u'', _(u'Language neutral')))
        return DisplayList(languages)

    #  DublinCore interface query methods #####################################

    security.declareProtected(permissions.View, 'CreationDate')

    def CreationDate(self, zone=None):
        # Dublin Core element - date resource created.
        if zone is None:
            zone = _zone
        creation = self.getField('creation_date').get(self)
        # return unknown if never set properly
        return creation is None and 'Unknown' or creation.toZone(zone).ISO8601()

    security.declareProtected(permissions.View, 'EffectiveDate')

    def EffectiveDate(self, zone=None):
        # Dublin Core element - date resource becomes effective.
        if zone is None:
            zone = _zone
        effective = self.getField('effectiveDate').get(self)
        return effective is None and 'None' or effective.toZone(zone).ISO8601()

    def _effective_date(self):
        # Computed attribute accessor.
        return self.getField('effectiveDate').get(self)

    security.declareProtected(permissions.View, 'effective_date')
    effective_date = ComputedAttribute(_effective_date, 1)

    security.declareProtected(permissions.View, 'ExpirationDate')

    def ExpirationDate(self, zone=None):
        # Dublin Core element - date resource expires.
        if zone is None:
            zone = _zone
        expires = self.getField('expirationDate').get(self)
        return expires is None and 'None' or expires.toZone(zone).ISO8601()

    def _expiration_date(self):
        # Computed attribute accessor.
        return self.getField('expirationDate').get(self)

    security.declareProtected(permissions.View, 'expiration_date')
    expiration_date = ComputedAttribute(_expiration_date, 1)

    security.declareProtected(permissions.View, 'Date')

    def Date(self, zone=None):
        # Dublin Core element - default date
        # Return effective_date if specifically set, modification date
        # otherwise
        if zone is None:
            zone = _zone
        effective = self.getField('effectiveDate').get(self)
        if effective is None:
            effective = self.modified()
        return (effective is None and DateTime().toZone(zone) or
                effective.toZone(zone).ISO8601())

    security.declareProtected(permissions.View, 'Format')

    def Format(self):
        # cmf/backward compat
        # Dublin Core element - resource format
        # FIXME: get content type from marshaller
        return self.getContentType()

    security.declareProtected(permissions.ModifyPortalContent,
                              'setFormat')

    def setFormat(self, value):
        # cmf/backward compat: ignore setFormat.
        self.setContentType(value)

    def Identifer(self):
        # dublin core getId method.
        return self.getId()

    #  DublinCore utility methods #############################################

    security.declareProtected(permissions.View, 'contentEffective')

    def contentEffective(self, date):
        # Is the date within the resource's effective range?
        effective = self.getField('effectiveDate').get(self)
        expires = self.getField('expirationDate').get(self)
        pastEffective = (effective is None or effective <= date)
        beforeExpiration = (expires is None or expires >= date)
        return pastEffective and beforeExpiration

    security.declareProtected(permissions.View, 'contentExpired')

    def contentExpired(self, date=None):
        # Is the date after resource's expiration.
        if not date:
            date = DateTime()
        expires = self.getField('expirationDate').get(self)
        if not expires:
            expires = CEILING_DATE
        return expires <= date

    #  CatalogableDublinCore methods ##########################################

    security.declareProtected(permissions.View, 'created')

    def created(self):
        # Dublin Core element - date resource created,
        # returned as DateTime.
        # allow for non-existent creation_date, existed always
        created = self.getField('creation_date').get(self)
        return created is None and FLOOR_DATE or created

    security.declareProtected(permissions.View, 'modified')

    def modified(self):
        # Dublin Core element - date resource last modified,
        # returned as DateTime.
        modified = self.getField('modification_date').get(self)
        # TODO may return None
        return modified

    security.declareProtected(permissions.View, 'effective')

    def effective(self):
        # Dublin Core element - date resource becomes effective,
        # returned as DateTime.
        effective = self.getField('effectiveDate').get(self)
        return effective is None and FLOOR_DATE or effective

    security.declareProtected(permissions.View, 'expires')

    def expires(self):
        # Dublin Core element - date resource expires,
        # returned as DateTime.
        expires = self.getField('expirationDate').get(self)
        return expires is None and CEILING_DATE or expires

    ## code below come from CMFDefault.DublinCore.DefaultDublinCoreImpl #######

    ###########################################################################
    #
    # Copyright (c) 2001 Zope Corporation and Contributors. All Rights Reserved
    #
    # This software is subject to the provisions of the Zope Public License,
    # Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
    # THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
    # WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    # WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
    # FOR A PARTICULAR PURPOSE
    #
    ###########################################################################

    #
    #  Set-modification-date-related methods.
    #  In DefaultDublinCoreImpl for lack of a better place.
    #

    security.declareProtected(permissions.ModifyPortalContent,
                              'notifyModified')

    def notifyModified(self):
        # Take appropriate action after the resource has been modified.
        # For now, change the modification_date.
        self.setModificationDate(DateTime())
        if shasattr(self, 'http__refreshEtag'):
            self.http__refreshEtag()

    security.declareProtected(permissions.ManagePortal,
                              'setModificationDate')

    def setModificationDate(self, modification_date=None):
        # Set the date when the resource was last modified.
        # When called without an argument, sets the date to now.
        if modification_date is None:
            modified = DateTime()
        else:
            modified = self._datify(modification_date)
        self.getField('modification_date').set(self, modified)

    security.declareProtected(permissions.ManagePortal,
                              'setCreationDate')

    def setCreationDate(self, creation_date=None):
        # Set the date when the resource was created.
        # When called without an argument, sets the date to now.
        if creation_date is None:
            created = DateTime()
        else:
            created = self._datify(creation_date)
        self.getField('creation_date').set(self, created)

    security.declarePrivate('_datify')

    def _datify(self, date):
        # Try to convert something into a DateTime instance or None.
        # stupid web
        if date == 'None':
            date = None
        if not isinstance(date, DateTime):
            if date is not None:
                date = DateTime(date)
        return date

    #
    #  DublinCore interface query methods
    #
    security.declareProtected(permissions.View, 'Publisher')

    def Publisher(self):
        # Dublin Core element - resource publisher
        # XXX: fixme using 'portal_metadata'
        return 'No publisher'

    security.declareProtected(permissions.View, 'ModificationDate')

    def ModificationDate(self, zone=None):
        # Dublin Core element - date resource last modified.
        if zone is None:
            zone = _zone
        modified = self.modified()
        return (modified is None and DateTime().toZone(zone)
                or modified.toZone(zone).ISO8601())

    security.declareProtected(permissions.View, 'Type')

    def Type(self):
        # Dublin Core element - Object type.
        if hasattr(aq_base(self), 'getTypeInfo'):
            ti = self.getTypeInfo()
            if ti is not None:
                return ti.Title()
        return self.meta_type

    security.declareProtected(permissions.View, 'Identifier')

    def Identifier(self):
        # Dublin Core element - Object ID
        # XXX: fixme using 'portal_metadata' (we need to prepend the
        #      right prefix to self.getPhysicalPath().
        return self.absolute_url()

    security.declareProtected(permissions.View, 'listContributors')

    def listContributors(self):
        # Dublin Core element - Contributors.
        return self.Contributors()

    security.declareProtected(permissions.ModifyPortalContent,
                              'addCreator')

    def addCreator(self, creator=None):
        # Add creator to Dublin Core creators.
        if creator is None:
            mtool = getToolByName(self, 'portal_membership', None)
            if mtool is None:
                return
            creator = mtool.getAuthenticatedMember().getId()

        # call self.listCreators() to make sure self.creators exists
        curr_creators = self.listCreators()
        if creator and not creator in curr_creators:
            self.setCreators(curr_creators + (creator, ))

    security.declareProtected(permissions.View, 'listCreators')

    def listCreators(self):
        # List Dublin Core Creator elements - resource authors.
        creators = self.Schema()['creators']
        if not creators.get(self):
            # for content created with CMF versions before 1.5
            owner_tuple = self.getOwnerTuple()
            owner_id = owner_tuple and owner_tuple[1]
            if owner_id:
                creators.set(self, (owner_id,))
            else:
                creators.set(self, ())

        return creators.get(self)

    security.declareProtected(permissions.View, 'Creator')

    def Creator(self):
        # Dublin Core Creator element - resource author.
        creators = self.listCreators()
        return creators and creators[0] or ''

    #
    #  DublinCore utility methods
    #

    # Deliberately *not* protected by a security declaration
    # See https://dev.plone.org/archetypes/ticket/712
    def content_type(self):
        """ WebDAV needs this to do the Right Thing (TM).
        """
        return self.Format()
    #
    #  CatalogableDublinCore methods
    #

    security.declareProtected(permissions.View, 'getMetadataHeaders')

    def getMetadataHeaders(self):
        # Return RFC-822-style headers.
        hdrlist = []
        hdrlist.append(('Title', self.Title()))
        hdrlist.append(('Subject', string.join(self.Subject(), ', ')))
        hdrlist.append(('Publisher', self.Publisher()))
        hdrlist.append(('Description', self.Description()))
        hdrlist.append(('Contributors', string.join(
            self.Contributors(), '; ')))
        hdrlist.append(('Creators', string.join(
            self.Creators(), '; ')))
        hdrlist.append(('Effective_date', self.EffectiveDate()))
        hdrlist.append(('Expiration_date', self.ExpirationDate()))
        hdrlist.append(('Type', self.Type()))
        hdrlist.append(('Format', self.Format()))
        hdrlist.append(('Language', self.Language()))
        hdrlist.append(('Rights', self.Rights()))
        return hdrlist

    #
    #  Management tab methods
    #

    security.declarePrivate('_editMetadata')

    def _editMetadata(self,
                      title=_marker,
                      subject=_marker,
                      description=_marker,
                      contributors=_marker,
                      effective_date=_marker,
                      expiration_date=_marker,
                      format=_marker,
                      language=_marker,
                      rights=_marker,
                      ):
        # Update the editable metadata for this resource.
        if title is not _marker:
            self.setTitle(title)
        if subject is not _marker:
            self.setSubject(subject)
        if description is not _marker:
            self.setDescription(description)
        if contributors is not _marker:
            self.setContributors(contributors)
        if effective_date is not _marker:
            self.setEffectiveDate(effective_date)
        if expiration_date is not _marker:
            self.setExpirationDate(expiration_date)
        if format is not _marker:
            self.setFormat(format)
        if language is not _marker:
            self.setLanguage(language)
        if rights is not _marker:
            self.setRights(rights)

    security.declareProtected(permissions.ModifyPortalContent,
                              'manage_metadata')
    manage_metadata = DTMLFile('zmi_metadata', config._www)

    security.declareProtected(permissions.ModifyPortalContent,
                              'manage_editMetadata')

    def manage_editMetadata(self,
                            title,
                            subject,
                            description,
                            contributors,
                            effective_date,
                            expiration_date,
                            format,
                            language,
                            rights,
                            REQUEST,
                            ):
        """ Update metadata from the ZMI.
        """
        self._editMetadata(title, subject, description, contributors,
                           effective_date, expiration_date,
                           format, language, rights,
                           )
        REQUEST['RESPONSE'].redirect(self.absolute_url()
                                     + '/manage_metadata'
                                     + '?manage_tabs_message=Metadata+updated.')

    security.declareProtected(permissions.ModifyPortalContent,
                              'editMetadata')

    def editMetadata(self,
                     title='',
                     subject=(),
                     description='',
                     contributors=(),
                     effective_date=None,
                     expiration_date=None,
                     format='text/html',
                     language='en-US',
                     rights='',
                     ):
        # Used to be:  editMetadata = WorkflowAction(_editMetadata)
        # Need to add check for webDAV locked resource for TTW methods.
        self.failIfLocked()
        self._editMetadata(title=title,
                           subject=subject,
                           description=description,
                           contributors=contributors,
                           effective_date=effective_date,
                           expiration_date=expiration_date,
                           format=format,
                           language=language,
                           rights=rights,
                           )
        self.reindexObject()

InitializeClass(ExtensibleMetadata)

ExtensibleMetadataSchema = ExtensibleMetadata.schema

__all__ = ('ExtensibleMetadata', 'ExtensibleMetadataSchema', )



from Products.CMFCore.utils import getToolByName

from Products.Archetypes.tests.attestcase import ATTestCase
from Products.Archetypes.tests.utils import makeContent


class RefSpeedupTestMixin:

    def afterSetUp(self):
        self.rc = getToolByName(self.portal, 'reference_catalog')
        self.loginAsPortalOwner()
        self.doc1 = makeContent(
            self.portal, portal_type='DDocument', id='doc1')
        self.doc2 = makeContent(
            self.portal, portal_type='DDocument', id='doc2')
        self.doc3 = makeContent(
            self.portal, portal_type='DDocument', id='doc3')


class TestGetReferences(RefSpeedupTestMixin, ATTestCase):

    def test_none(self):
        self.assertEqual(self.rc.getReferences(self.doc1), [])
        self.assertEqual(self.rc.getReferences(self.doc1, 'related'), [])
        self.assertEqual(self.rc.getReferences(
            self.doc1, 'related', self.doc2), [])

        result = self.rc.getReferences(self.doc1, ['related', 'rel2'])
        self.assertEqual(result, [])
        result = self.rc.getReferences(
            self.doc1, ['related', 'rel2'], self.doc2)
        self.assertEqual(result, [])

        self.assertEqual(self.doc1.getReferences(), [])

    def test_single(self):
        self.doc1.setRelated([self.doc2.UID()])
        result = self.rc.getReferences(self.doc1)
        self.assertEqual(result[0].getTargetObject(), self.doc2)
        result = self.rc.getReferences(self.doc1, 'related')
        self.assertEqual(result[0].getTargetObject(), self.doc2)
        result = self.rc.getReferences(self.doc1, 'related', self.doc2)
        self.assertEqual(result[0].getTargetObject(), self.doc2)
        result = self.rc.getReferences(self.doc1, 'related', self.doc3)
        self.assertEqual(result, [])

        self.assertEqual(self.doc1.getReferences()[0], self.doc2)

        self.doc1.setRel2([self.doc3.UID()])
        result = self.rc.getReferences(self.doc1, ['related', 'rel2'])
        result = [r.getTargetObject() for r in result]
        self.assertEqual(set(result), set([self.doc2, self.doc3]))

        result = self.rc.getReferences(
            self.doc1, ['related', 'rel2'], self.doc2)
        result = [r.getTargetObject() for r in result]
        self.assertEqual(set(result), set([self.doc2]))

    def test_many(self):
        uids = [self.doc2.UID(), self.doc3.UID()]
        self.doc1.setRelated(uids)
        result = [r.getTargetObject()
                  for r in self.rc.getReferences(self.doc1)]
        self.assertEqual(set(result), set([self.doc2, self.doc3]))

        self.assertEqual(set(self.doc1.getReferences()),
                         set([self.doc2, self.doc3]))

        self.doc1.setRel2([self.doc2.UID()])
        result = self.rc.getReferences(self.doc1, ['related', 'rel2'])
        result = [r.getTargetObject() for r in result]
        self.assertEqual(set(result), set([self.doc2, self.doc3]))

    def test_bidi(self):
        self.doc1.setRelated([self.doc2.UID()])
        self.doc2.setRelated([self.doc1.UID()])
        result = [r.getTargetObject()
                  for r in self.rc.getReferences(self.doc1)]
        self.assertEqual(result, [self.doc2])
        result = [r.getTargetObject()
                  for r in self.rc.getReferences(self.doc2)]
        self.assertEqual(result, [self.doc1])

        self.doc1.setRel2([self.doc3.UID()])
        result = self.rc.getReferences(self.doc1, ['related', 'rel2'])
        result = [r.getTargetObject() for r in result]
        self.assertEqual(set(result), set([self.doc2, self.doc3]))

    def test_missing_uid_catalog_entry(self):
        self.doc1.setRelated([self.doc2.UID()])

        result = [r.getTargetObject()
                  for r in self.rc.getReferences(self.doc1)]
        self.assertEqual(result, [self.doc2])

        # Forcefully remove the target object from the uid catalog
        uc = getToolByName(self.portal, 'uid_catalog')
        uc.uncatalog_object(self.doc2._getURL())

        references = self.rc.getReferences(self.doc1)
        self.assertEqual(len(references), 1)
        self.assertEqual(references[0].getTargetObject(), None)


class TestGetBackReferences(RefSpeedupTestMixin, ATTestCase):

    def test_none(self):
        self.assertEqual(self.rc.getBackReferences(self.doc1), [])
        self.assertEqual(self.rc.getBackReferences(self.doc1, 'related'), [])
        self.assertEqual(self.rc.getBackReferences(self.doc1, 'related', self.doc2),
                         [])

        result = self.rc.getBackReferences(self.doc1, ['related', 'rel2'])
        self.assertEqual(result, [])
        result = self.rc.getBackReferences(
            self.doc1, ['related', 'rel2'], self.doc2)
        self.assertEqual(result, [])

        self.assertEqual(self.doc1.getBackReferences(), [])

    def test_single(self):
        self.doc1.setRelated([self.doc2.UID()])

        result = self.rc.getBackReferences(self.doc2)
        self.assertEqual(result[0].getSourceObject(), self.doc1)
        result = self.rc.getBackReferences(self.doc2, 'related')
        self.assertEqual(result[0].getSourceObject(), self.doc1)
        result = self.rc.getBackReferences(self.doc2, 'related', self.doc1)
        self.assertEqual(result[0].getSourceObject(), self.doc1)
        result = self.rc.getBackReferences(self.doc2, 'related', self.doc3)
        self.assertEqual(result, [])

        self.assertEqual(self.doc2.getBackReferences('related')[0], self.doc1)

        self.doc1.setRel2([self.doc3.UID()])
        result = self.rc.getBackReferences(self.doc2, ['related', 'rel2'])
        self.assertEqual(result[0].getSourceObject(), self.doc1)
        result = self.rc.getBackReferences(
            self.doc2, ['related', 'rel2'], self.doc1)
        self.assertEqual(result[0].getSourceObject(), self.doc1)

    def test_many(self):
        uids = [self.doc2.UID(), self.doc3.UID()]
        self.doc1.setRelated(uids)

        result = [r.getSourceObject()
                  for r in self.rc.getBackReferences(self.doc2)]
        self.assertEqual(set(result), set([self.doc1]))
        result = [r.getSourceObject()
                  for r in self.rc.getBackReferences(self.doc3)]
        self.assertEqual(set(result), set([self.doc1]))

        self.assertEqual(set(self.doc2.getBackReferences()), set([self.doc1]))

        uids2 = [self.doc1.UID(), self.doc2.UID()]
        self.doc3.setRel2(uids2)
        result = self.rc.getBackReferences(self.doc2, ['related', 'rel2'])
        result = [r.getSourceObject() for r in result]
        self.assertEqual(set(result), set([self.doc1, self.doc3]))

    def test_bidi(self):
        self.doc1.setRelated([self.doc2.UID()])
        self.doc2.setRelated([self.doc1.UID()])

        result = [r.getSourceObject()
                  for r in self.rc.getBackReferences(self.doc1)]
        self.assertEqual(result, [self.doc2])
        result = [r.getSourceObject()
                  for r in self.rc.getBackReferences(self.doc2)]
        self.assertEqual(result, [self.doc1])

        self.doc2.setRel2([self.doc1.UID()])
        result = self.rc.getBackReferences(self.doc2, ['related', 'rel2'])
        result = [r.getSourceObject() for r in result]
        self.assertEqual(set(result), set([self.doc1]))

    def test_missing_uid_catalog_entry(self):
        self.doc2.setRelated([self.doc1.UID()])

        result = [r.getSourceObject()
                  for r in self.rc.getBackReferences(self.doc1)]
        self.assertEqual(result, [self.doc2])

        # Forcefully remove the target object from the uid catalog
        uc = getToolByName(self.portal, 'uid_catalog')
        uc.uncatalog_object(self.doc2._getURL())

        references = self.rc.getBackReferences(self.doc1)
        self.assertEqual(len(references), 1)
        self.assertEqual(references[0].getSourceObject(), None)


class TestReferenceable(RefSpeedupTestMixin, ATTestCase):

    def test_no_references(self):
        self.assertEqual(self.doc1.getRelated(), [])
        self.assertEqual(self.doc1.getRawRelated(), [])
        self.assertEqual(self.doc1.getReferences(), [])
        self.assertEqual(self.doc1.getRelationships(), [])

    def test_single_reference(self):
        self.doc1.setRelated([self.doc2.UID()])
        self.assertEqual(self.doc1.getRelated(), [self.doc2])
        self.assertEqual(self.doc1.getRawRelated(), [self.doc2.UID()])
        self.assertEqual(self.doc1.getReferences(), [self.doc2])
        self.assertEqual(self.doc1.getRelationships(), ['related'])

    def test_many_references(self):
        uids = [self.doc2.UID(), self.doc3.UID()]
        self.doc1.setRelated(uids)
        self.assertEqual(set(self.doc1.getRelated()),
                         set([self.doc2, self.doc3]))
        self.assertEqual(set(self.doc1.getRawRelated()), set(uids))
        self.assertEqual(set(self.doc1.getReferences()),
                         set([self.doc2, self.doc3]))
        self.assertEqual(self.doc1.getRelationships(), ['related'])

    def test_bidi_references(self):
        self.doc1.setRelated([self.doc2.UID()])
        self.doc2.setRelated([self.doc1.UID()])
        self.assertEqual(self.doc1.getRelated(), [self.doc2])
        self.assertEqual(self.doc1.getRawRelated(), [self.doc2.UID()])
        self.assertEqual(self.doc1.getReferences(), [self.doc2])
        self.assertEqual(self.doc1.getRelationships(), ['related'])
        self.assertEqual(self.doc2.getRelated(), [self.doc1])
        self.assertEqual(self.doc2.getRawRelated(), [self.doc1.UID()])
        self.assertEqual(self.doc2.getReferences(), [self.doc1])
        self.assertEqual(self.doc2.getRelationships(), ['related'])



from zope.interface import Interface


class ILayer(Interface):
    """Layering support
    """

    def initializeInstance(instance, item=None, container=None):
        """Optionally called to initialize a layer for an entire
        instance
        """

    def initializeField(instance, field):
        """Optionally called to initialize a layer for a given field
        """

    def cleanupField(instance, field):
        """Optionally called to cleanup a layer for a given field
        """

    def cleanupInstance(instance, item=None, container=None):
        """Optionally called to cleanup a layer for an entire
        instance
        """


class ILayerContainer(Interface):
    """An object that contains layers and can use/manipulate them"""

    def registerLayer(name, object):
        """Register an object as providing a new layer under a given
        name
        """

    def registeredLayers():
        """Provides a list of (name, object) layer pairs
        """

    def hasLayer(name):
        """Boolean indicating if the layer is implemented on the
        object
        """

    def getLayerImpl(name):
        """Return an object implementing this layer
        """


class ILayerRuntime(Interface):
    """ Layer Runtime """

    def initializeLayers(instance, item=None, container=None):
        """Optionally process all layers attempting their
        initializeInstance and initializeField methods if they exist
        """

    def cleanupLayers(instance, item=None, container=None):
        """Optionally process all layers attempting their
        cleanupInstance and cleanupField methods if they exist.
        """



from zope.event import notify

from Acquisition import aq_inner

from Products.Five.browser import BrowserView

from Products.Archetypes.event import EditBegunEvent
from Products.Archetypes.event import EditCancelledEvent


class Lifecycle(BrowserView):
    """Helper functions to trigger lifecycle events from TTW code
    """

    def begin_edit(self):
        """Begin an edit operation
        """
        notify(EditBegunEvent(aq_inner(self.context)))

    def cancel_edit(self):
        """Cancel an edit operation
        """
        notify(EditCancelledEvent(aq_inner(self.context)))



##########################################################################
#
# Copyright (c) 2002-2005, Benjamin Saller <bcsaller@ideasuite.com>, and
#                              the respective authors. All rights reserved.
# For a list of Archetypes contributors see docs/CREDITS.txt.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the author nor the names of its contributors may be used
#   to endorse or promote products derived from this software without specific
#   prior written permission.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##########################################################################

from Acquisition import aq_base
from AccessControl import ClassSecurityInfo

from Products.Archetypes.Storage import Storage
from Products.Archetypes.Storage import StorageLayer
from Products.Archetypes.Storage import _marker
from Products.Archetypes.annotations import AT_ANN_STORAGE
from Products.Archetypes.annotations import AT_MD_STORAGE
from Products.Archetypes.annotations import getAnnotation
from Products.Archetypes.Registry import setSecurity
from Products.Archetypes.Registry import registerStorage
from Products.Archetypes.utils import shasattr
from Products.CMFCore.utils import getToolByName


class BaseAnnotationStorage(Storage):
    """Stores data using annotations on the instance
    """

    security = ClassSecurityInfo()

    _key = None

    def __init__(self, migrate=False):
        self._migrate = migrate

    def _migration(self, name, instance, **kwargs):
        """Migrates data from the original storage
        """
        raise NotImplementedError

    def _cleanup(self, name, instance, value, **kwargs):
        """Clean up data in set method
        """
        raise NotImplementedError

    security.declarePrivate('get')

    def get(self, name, instance, **kwargs):
        ann = getAnnotation(instance)
        value = ann.getSubkey(self._key, subkey=name, default=_marker)
        if value is _marker:
            if self._migrate:
                value = self._migration(name, instance, **kwargs)
            else:
                raise AttributeError(name)
        return value

    security.declarePrivate('set')

    def set(self, name, instance, value, **kwargs):
        # Remove acquisition wrappers
        value = aq_base(value)
        ann = getAnnotation(instance)
        ann.setSubkey(self._key, value, subkey=name)
        if self._migrate:
            self._cleanup(name, instance, value, **kwargs)

    security.declarePrivate('unset')

    def unset(self, name, instance, **kwargs):
        ann = getAnnotation(instance)
        try:
            ann.delSubkey(self._key, subkey=name)
        except KeyError:
            pass

setSecurity(BaseAnnotationStorage)


class AnnotationStorage(BaseAnnotationStorage):
    """Stores values as ATAnnotations on the object
    """

    _key = AT_ANN_STORAGE

    security = ClassSecurityInfo()

    def _migration(self, name, instance, **kwargs):
        """Migrates data from the original storage
        """
        value = getattr(aq_base(instance), name, _marker)
        if value is _marker:
            raise AttributeError(name)
        # explicit del althought set would do the job, too
        delattr(instance, name)
        self.set(name, instance, value, **kwargs)
        return value

    def _cleanup(self, name, instance, value, **kwargs):
        if shasattr(instance, name):
            delattr(instance, name)

registerStorage(AnnotationStorage)


class MetadataAnnotationStorage(BaseAnnotationStorage, StorageLayer):
    """Stores metadata as ATAnnotations on the object
    """

    _key = AT_MD_STORAGE

    security = ClassSecurityInfo()

    def _migration(self, name, instance, **kwargs):
        """Migrates data from the original storage
        """
        try:
            md = aq_base(instance)._md
            value = md[name]
        except KeyError, msg:
            # We are acting like an attribute, so
            # raise AttributeError instead of KeyError
            raise AttributeError(name, msg)
        self.set(name, instance, value, **kwargs)
        del md[name]
        return value

    def _cleanup(self, name, instance, value, **kwargs):
        md = aq_base(instance)._md
        try:
            del md[name]
        except KeyError:
            pass

    security.declarePrivate('initializeInstance')

    def initializeInstance(self, instance, item=None, container=None):
        # annotations are initialized on first access
        pass

    security.declarePrivate('initializeField')

    def initializeField(self, instance, field):
        # Check for already existing field to avoid  the reinitialization
        # (which means overwriting) of an already existing field after a
        # copy or rename operation
        ann = getAnnotation(instance)
        if not ann.hasSubkey(self._key, subkey=field.getName()):
            self.set(field.getName(), instance, field.getDefault(instance))

    security.declarePrivate('cleanupField')

    def cleanupField(self, instance, field, **kwargs):
        # Don't clean up the field self to avoid problems with copy/rename. The
        # python garbarage system will clean up if needed.
        pass

    security.declarePrivate('cleanupInstance')

    def cleanupInstance(self, instance, item=None, container=None):
        # Don't clean up the instance self to avoid problems with copy/rename. The
        # python garbarage system will clean up if needed.
        pass

registerStorage(MetadataAnnotationStorage)


def migrateStorageOfType(portal, portal_type, schema):
    """Migrate storage from attribute to annotation storage

    portal - portal
    portal_type - portal type name to migrate
    schema - schema of the type

    The schema is used to detect annotation and metadata annotation stored field for
    migration.
    """
    catalog = getToolByName(portal, 'portal_catalog')
    brains = catalog(dict(Type=portal_type))

    fields = [field.getName()
              for field in schema.fields()
              if field.storage.__class__ == AnnotationStorage
              ]
    md_fields = [field.getName()
                 for field in schema.fields()
                 if field.storage.__class__ == MetadataAnnotationStorage
                 ]

    for brain in brains:
        obj = brain.getObject()
        if obj is None:
            continue

        try:
            state = obj._p_changed
        except:
            state = 0

        ann = getAnnotation(obj)
        clean_obj = aq_base(obj)
        _attr2ann(clean_obj, ann, fields)
        _meta2ann(clean_obj, ann, md_fields)

        if state is None:
            obj._p_deactivate()


def _attr2ann(clean_obj, ann, fields):
    """Attribute 2 annotation
    """
    for field in fields:
        if not ann.hasSubkey(AT_ANN_STORAGE, field):
            value = getattr(clean_obj, field, _marker)
            if value is not _marker:
                delattr(clean_obj, field)
                ann.setSubkey(AT_ANN_STORAGE, value, subkey=field)
        else:
            value = getattr(clean_obj, field, _marker)
            if value is not _marker:
                delattr(clean_obj, field)


def _meta2ann(clean_obj, ann, fields):
    """metadata 2 annotation
    """
    md = clean_obj._md
    for field in fields:
        if not ann.hasSubkey(AT_MD_STORAGE, field):
            value = md.get(field, _marker)
            if value is not _marker:
                del md[field]
                ann.setSubkey(AT_MD_STORAGE, value, subkey=field)
        else:
            try:
                del md[field]
            except KeyError:
                pass



from zope.component import adapts

from Products.CMFCore.utils import getToolByName
from Products.GenericSetup.interfaces import ISetupEnviron
from Products.GenericSetup.utils import exportObjects
from Products.GenericSetup.utils import importObjects
from Products.GenericSetup.utils import XMLAdapterBase


from Products.Archetypes.config import TOOL_NAME
from Products.Archetypes.interfaces import IArchetypeTool


class ArchetypeToolXMLAdapter(XMLAdapterBase):
    """Mode in- and exporter for ArchetypesTool.
    """

    adapts(IArchetypeTool, ISetupEnviron)

    def _exportNode(self):
        """Export the object as a DOM node.
        """
        node = self._doc.createElement('archetypetool')
        node.appendChild(self._extractCatalogSettings())

        self._logger.info('ArchetypeTool settings exported.')
        return node

    def _importNode(self, node):
        if self.environ.shouldPurge():
            self._purgeCatalogSettings()

        self._initCatalogSettings(node)
        self._logger.info('ArchetypeTool settings imported.')

    def _purgeCatalogSettings(self):
        self.context.catalog_map.clear()

    def _initCatalogSettings(self, node):
        for child in node.childNodes:
            if child.nodeName == 'catalogmap':
                for type in child.getElementsByTagName('type'):
                    portaltype = type.getAttribute('portal_type')
                    catalogs = [e.getAttribute('value')
                                for e in type.getElementsByTagName('catalog')]
                    already = [cat.getId() for cat in
                               self.context.getCatalogsByType(portaltype)]
                    catalogs = set(catalogs + already)
                    self.context.setCatalogsByType(portaltype, list(catalogs))

    def _extractCatalogSettings(self):
        node = self._doc.createElement('catalogmap')
        for type in self.context.listRegisteredTypes(True):
            child = self._doc.createElement('type')
            child.setAttribute('portal_type', type['name'])
            for cat in self.context.getCatalogsByType(type['name']):
                sub = self._doc.createElement('catalog')
                sub.setAttribute('value', cat.id)
                child.appendChild(sub)
            node.appendChild(child)

        return node


def importArchetypeTool(context):
    """Import Archetype Tool configuration.
    """
    site = context.getSite()
    logger = context.getLogger("archetypetool")
    tool = getToolByName(site, TOOL_NAME, None)
    if tool is None:
        return

    importObjects(tool, '', context)
    logger.info("Archetype tool imported.")


def exportArchetypeTool(context):
    """Export Archetype Tool configuration.
    """
    site = context.getSite()
    logger = context.getLogger("archetypetool")
    tool = getToolByName(site, TOOL_NAME, None)
    if tool is None:
        return

    exportObjects(tool, '', context)
    logger.info("Archetype tool exported.")



from Products.Archetypes.atapi import *
from DateTime import DateTime
from Products.Archetypes.config import PKG_NAME

schema = BaseSchema.copy() + Schema((
    TextField('quote',
              searchable=1,
              required=1,
              ),

    LinesField('sources',
               widget=LinesWidget,
               ),

    TextField('footnote',
              required=1,
              widget=TextAreaWidget,
              ),

    DateTimeField('fact_date',
                  default_method=DateTime,
                  widget=CalendarWidget(label="Date"),
                  ),

    StringField('url',
                widget=StringWidget(description="A URL citing the fact",
                                    label="URL"),
                validators=('isURL',),
                ),
))


class Fact(BaseContent):
    """A quoteable fact or tidbit"""
    schema = schema


registerType(Fact, PKG_NAME)



from zope.interface import implements

from Acquisition import aq_base, aq_inner, aq_parent
from OFS.ObjectManager import BeforeDeleteException
from ZODB.POSException import ConflictError

from Products.Archetypes.config import TOOL_NAME, MYSQL_SQLSTORAGE_TABLE_TYPE
from Products.Archetypes.interfaces.field import IObjectField
from Products.Archetypes.interfaces.layer import ILayer
from Products.Archetypes.interfaces.storage import ISQLStorage
from Products.Archetypes.log import log
from Products.Archetypes.SQLMethod import SQLMethod
from Products.Archetypes.Storage import StorageLayer, type_map
from Products.CMFCore.utils import getToolByName


class BaseSQLStorage(StorageLayer):
    """ SQLStorage Base, more or less ISO SQL """

    implements(ISQLStorage, ILayer)

    query_create = ('create table <dtml-var table> '
                    '(UID char(50) primary key not null, '
                    'PARENTUID char(50), <dtml-var columns>)')
    query_drop = ('drop table <dtml-var table>')
    query_select = ('select <dtml-var field> from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')
    query_insert = ('insert into <dtml-var table> '
                    'set UID=<dtml-sqlvar UID type="string">, '
                    'PARENTUID=<dtml-sqlvar PARENTUID type="string">')
    query_update = ('update <dtml-var table> set '
                    '<dtml-var field>=<dtml-sqlvar value '
                    'type="%s" optional> where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_delete = ('delete from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')

    sqlm_type_map = {'integer': 'int'}

    db_type_map = {'fixedpoint': 'integer'}

    def map_object(self, field, value):
        if value is None:
            return 'None'
        else:
            return value

    def unmap_object(self, field, value):
        if value == 'None':
            return None
        else:
            return value

    def map_datetime(self, field, value):
        # we don't want to lose even 0.001 second
        try:
            return value.ISO()[:-2] + str(value.second())
        except:
            return None

    def map_fixedpoint(self, field, value):
        __traceback_info__ = repr(value)
        template = '%%d%%0%dd' % field.precision
        return template % value

    def unmap_fixedpoint(self, field, value):
        __traceback_info__ = repr(value)
        if value is None or value == '':
            return (0, 0)
        if type(value) == type(''):   # Gadfly return integers as strings
            value = int(value)
        split = 10 ** field.precision
        return (value / split), (value % split)

    def map_lines(self, field, value):
        __traceback_info__ = repr(value)
        return '\n'.join(value)

    def unmap_lines(self, field, value):
        __traceback_info__ = repr(value)
        return value.split('\n')

    def map_boolean(self, field, value):
        __traceback_info__ = repr(value)
        if not value:
            return 0
        else:
            return 1

    def map_reference(self, field, value):
        __traceback_info__ = repr(value)

        return ','.join(value)

    def unmap_boolean(self, field, value):
        __traceback_info__ = repr(value)
        if not value or value == '0':   # Gadfly return integers as strings
            return 0
        else:
            return 1

    def table_exists(self, instance):
        raise NotImplemented

    def is_initialized(self, instance):
        try:
            return self.getName() in instance.__initialized
        except AttributeError:
            return None

    def initializeField(self, instance, field):
        pass

    def is_cleaned(self, instance):
        try:
            return self.getName() in instance.__cleaned
        except AttributeError:
            return None

    def cleanupField(self, instance, field):
        pass

    def _query(self, instance, query, args):
        c_tool = getToolByName(instance, TOOL_NAME)
        connection_id = c_tool.getConnFor(instance)
        method = SQLMethod(instance)
        method.edit(connection_id, ' '.join(args.keys()), query)
        query, result = method(test__=1, **args)
        return result

    def initializeInstance(self, instance, item=None, container=None):
        if (self.is_initialized(instance) or
                getattr(instance, '_at_is_fake_instance', None)):
            # duh, we don't need to be initialized twice
            return
        factory = getToolByName(instance, 'portal_factory')
        if factory.isTemporary(instance):
            return

        fields = instance.Schema().fields()
        fields = [f for f in fields if IObjectField.providedBy(f)
                  and f.getStorage().__class__ is self.__class__]
        columns = []
        args = {}
        for field in fields:
            type = self.db_type_map.get(field.type, field.type)
            name = field.getName()
            # MySQL supports escape for columns names!
            if self.__class__.__name__ == 'MySQLSQLStorage':
                columns.append('`%s` %s' % (name, type))
            else:
                columns.append('%s %s' % (name, type))
        parent = container or aq_parent(aq_inner(instance))
        args['PARENTUID'] = getattr(aq_base(parent), 'UID', lambda: None)()
        args['table'] = instance.portal_type
        args['UID'] = instance.UID()
        # args['db_encoding']=kwargs.get('db_encoding',None)
        args['columns'] = ', ' + ', '.join(columns)
        if not self.table_exists(instance):
            self._query(instance, self.query_create, args)
            log('created table %s\n' % args['table'])
        try:
            self._query(instance, self.query_insert, args)
        except ConflictError:
            raise
        except:
            # usually, duplicate key
            # raise SQLInitException(msg)
            raise
        try:
            instance.__initialized += (self.getName(),)
        except AttributeError:
            instance.__initialized = (self.getName(),)
        # now, if we find an attribute called _v_$classname_temps, it
        # means the object was moved and we can initialize the fields
        # with those values
        temps_var = '_v_%s_temps' % self.getName()
        if hasattr(aq_base(instance), temps_var):
            temps = getattr(instance, temps_var)
            for key, value in temps.items():
                instance.Schema()[key].set(instance, value)
            delattr(instance, temps_var)
        try:
            del instance.__cleaned
        except (AttributeError, KeyError):
            pass

    def get(self, name, instance, **kwargs):
        if not self.is_initialized(instance):
            # ignore all calls before we're initialized - some
            # manage_afterAdd() methods try to get and set fields and
            # we can't allow that to break
            return None
        field = kwargs.get('field', instance.getField(name))
        args = {}
        args['table'] = instance.portal_type
        args['UID'] = instance.UID()
        args['db_encoding'] = kwargs.get('db_encoding', None)
        args['field'] = name
        result = self._query(instance, self.query_select, args)
        result = result[0][0]
        mapper = getattr(self, 'unmap_' + field.type, None)
        if mapper is not None:
            result = mapper(field, result)
        return result

    def set(self, name, instance, value, **kwargs):
        if not self.is_initialized(instance):
            # ignore all calls before we're initialized - some
            # manage_afterAdd() methods try to get and set fields and
            # we can't allow that to break
            return None
        field = kwargs.get('field', instance.getField(name))
        mapper = getattr(self, 'map_' + field.type, None)
        if mapper is not None:
            value = mapper(field, value)
        type = type_map.get(field.type, 'string')
        sql_type = self.sqlm_type_map.get(field.type, 'string')
        default = field.default
        args = {}
        args['table'] = instance.portal_type
        args['UID'] = instance.UID()
        # args['db_encoding']=kwargs.get('db_encoding',None)
        field_name = '%s:%s' % (name, type)
        if default:
            if type == 'string':
                default = "'%s'" % default
            field_name = "%s=%s" % (name, default)
        args[field_name] = name
        args['field'] = name
        if value is not None:
            # omiting it causes dtml-sqlvar to insert NULL
            args['value'] = value
        self._query(instance, self.query_update % sql_type, args)

    def cleanupInstance(self, instance, item=None, container=None):
        if (self.is_cleaned(instance) or
                getattr(instance, '_at_is_fake_instance', None)):
            # duh, we don't need to be cleaned twice
            return
        # the object is being deleted. remove data from sql.  but
        # first, made a temporary copy of the field values in case we
        # are being moved
        fields = instance.Schema().fields()
        fields = [f for f in fields if IObjectField.providedBy(f)
                  and f.getStorage().__class__ is self.__class__]
        temps = {}
        for f in fields:
            temps[f.getName()] = f.get(instance)
        setattr(instance, '_v_%s_temps' % self.getName(), temps)
        # now, remove data from sql
        c_tool = getToolByName(instance, TOOL_NAME)
        connection_id = c_tool.getConnFor(instance)
        args = {}
        args['table'] = instance.portal_type
        args['UID'] = instance.UID()
        # args['db_encoding']=kwargs.get('db_encoding',None)
        method = SQLMethod(instance)
        method.edit(connection_id, ' '.join(args.keys()), self.query_delete)
        try:
            query, result = method(test__=1, **args)
        except ConflictError:
            raise
        except:
            # dunno what could happen here raise
            # SQLCleanupException(msg)
            raise BeforeDeleteException
        try:
            instance.__cleaned += (self.getName(),)
        except AttributeError:
            instance.__cleaned = (self.getName(),)
        try:
            del instance.__initialized
        except (AttributeError, KeyError):
            pass


class GadflySQLStorage(BaseSQLStorage):

    query_create = ('create table <dtml-var table> '
                    '(UID varchar, PARENTUID varchar <dtml-var columns>)')
    query_select = ('select <dtml-var field> from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')
    query_insert = ('insert into <dtml-var table> '
                    'values (<dtml-sqlvar UID type="string">, '
                    '<dtml-sqlvar PARENTUID type="string">, '
                    '<dtml-in expr="_.string.split(columns,\',\')[1:]"> '
                    '<dtml-if sequence-end>\'\'<dtml-else>\'\', </dtml-if> '
                    '</dtml-in>) ')
    query_update = ('update <dtml-var table> set '
                    '<dtml-var field>=<dtml-sqlvar value '
                    'type="%s" optional> where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_delete = ('delete from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')

    sqlm_type_map = {'integer': 'string',
                     'float': 'string'}

    db_type_map = {'object': 'varchar',
                   'string': 'varchar',
                   'text': 'varchar',
                   'datetime': 'varchar',
                   'integer': 'varchar',
                   'float': 'varchar',
                   'fixedpoint': 'integer',
                   'lines': 'varchar',
                   'reference': 'varchar',
                   'boolean': 'integer',
                   }

    def map_datetime(self, field, value):
        try:
            return value.ISO()[:-2] + str(value.second())
        except:
            return ''

    def unmap_datetime(self, field, value):
        from DateTime import DateTime
        try:
            return DateTime(value)
        except:
            return None

    def map_integer(self, field, value):
        __traceback_info__ = repr(value)
        if value is None:   # Gadfly represents None as an empty string
            return ''
        else:
            return str(value)

    def unmap_integer(self, field, value):
        __traceback_info__ = repr(value)
        if value == '':   # Gadfly represents None as an empty string
            return None
        else:
            return int(value)

    def map_float(self, field, value):
        __traceback_info__ = repr(value)
        if value is None:   # Gadfly represents None as an empty string
            return ''
        else:
            return str(value)

    def unmap_float(self, field, value):
        __traceback_info__ = repr(value)
        if value == '':   # Gadfly represents None as an empty string
            return None
        else:
            return float(value)

    def unmap_lines(self, field, value):
        __traceback_info__ = repr(value)
        if value == '':   # Gadfly represents None as an empty String
            return None
        else:
            return value.split('\n')

    def table_exists(self, instance):
        try:
            self._query(instance,
                        'select * from <dtml-var table>',
                        {'table': instance.portal_type.lower()})
        except ConflictError:
            raise
        except:
            return 0
        else:
            return 1


class MySQLSQLStorage(BaseSQLStorage):

    query_create = ('create table `<dtml-var table>` '
                    '(UID char(50) primary key not null, '
                    'PARENTUID char(50) <dtml-var columns>) TYPE = %s' % MYSQL_SQLSTORAGE_TABLE_TYPE)
    query_select = ('select `<dtml-var field>` '
                    'from `<dtml-var table>` where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_insert = ('insert into `<dtml-var table>` '
                    'set UID=<dtml-sqlvar UID type="string">, '
                    'PARENTUID=<dtml-sqlvar PARENTUID type="string">')
    query_update = ('update `<dtml-var table>` set '
                    '`<dtml-var field>`=<dtml-sqlvar value '
                    'type="%s" optional> where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_delete = ('delete from `<dtml-var table>` '
                    'where <dtml-sqltest UID op="eq" type="string">')

    db_type_map = {'object': 'text',
                   'string': 'text',
                   'fixedpoint': 'integer',
                   'lines': 'text',
                   'reference': 'text',
                   'boolean': 'tinyint',
                   }

    def table_exists(self, instance):
        result = [r[0].lower() for r in
                  self._query(instance, '''show tables''', {})]
        return instance.portal_type.lower() in result


class PostgreSQLStorage(BaseSQLStorage):

    query_create = ('create table <dtml-var table> '
                    '(UID text primary key not null, '
                    'PARENTUID text <dtml-var columns>)')
    query_select = ('select <dtml-var field> from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')
    query_insert = ('insert into <dtml-var table> '
                    '(UID, PARENTUID) values '
                    '(<dtml-sqlvar UID type="string">, '
                    '<dtml-sqlvar PARENTUID type="string">)')
    query_update = ('update <dtml-var table> set '
                    '<dtml-var field>=<dtml-sqlvar value '
                    'type="%s" optional> where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_delete = ('delete from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')

    db_type_map = {'object': 'text',
                   'string': 'text',
                   'datetime': 'timestamp',
                   'fixedpoint': 'integer',
                   'lines': 'text',
                   'reference': 'text',
                   }

    def table_exists(self, instance):
        return self._query(instance,
                           ('select relname from pg_class where '
                            '<dtml-sqltest relname op="eq" type="string">'),
                           {'relname': instance.portal_type.lower()})


class SQLServerStorage(BaseSQLStorage):

    query_create = ('create table <dtml-var table> '
                    '(UID varchar(50) CONSTRAINT pk_uid '
                    'PRIMARY KEY CLUSTERED, '
                    'PARENTUID varchar(50) '
                    '<dtml-var columns>)')
    query_select = ('select <dtml-var field> from '
                    '<dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')
    query_insert = ('insert into <dtml-var table> '
                    '(UID, PARENTUID) values '
                    '(<dtml-sqlvar UID type="string">, '
                    '<dtml-sqlvar PARENTUID type="string">)')
    query_update = ('update <dtml-var table> set '
                    '<dtml-var field>=<dtml-sqlvar value '
                    'type="%s" optional> where '
                    '<dtml-sqltest UID op="eq" type="string">')
    query_delete = ('delete from <dtml-var table> '
                    'where <dtml-sqltest UID op="eq" type="string">')

    db_type_map = {'object': 'varchar',
                   'string': 'varchar',
                   'text': 'varchar',
                   'datetime': 'timestamp',
                   'fixedpoint': 'integer',
                   'lines': 'varchar',
                   'reference': 'varchar',
                   'boolean': 'integer',
                   }

    def table_exists(self, instance):
        return self._query(instance,
                           ('select name from '
                            'sysobjects where '
                            'xtype=char(85) and uid=1 and '
                            '<dtml-sqltest name op="eq" type="string">'),
                           {'name': instance.portal_type.lower()})
